import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="626",
    message="Child was reported as a mother but the date of birth of the first child is before the current "
    + "year which contradicts with the mother status recorded last year.",
    affected_fields=["MOTHER", "MC_DOB"],
    tables=["Header", "Header_last"],
)
def validate(dfs):
    if "Header" not in dfs or "Header_last" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        header_prev = dfs["Header_last"]
        collection_start = dfs["metadata"]["collection_start"]
        header["MC_DOB"] = pd.to_datetime(
            header["MC_DOB"], format="%d/%m/%Y", errors="coerce"
        )
        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )
        header["orig_idx"] = header.index
        header = header.query("MC_DOB.notna()")
        merged = header.merge(
            header_prev, how="inner", on="CHILD", suffixes=["", "_PRE"]
        )
        merged["MOTHER"] = pd.to_numeric(merged["MOTHER"], errors="coerce")
        merged["MOTHER_PRE"] = pd.to_numeric(merged["MOTHER_PRE"], errors="coerce")
        err_co = merged[
            (merged["MOTHER"] == 1)
            & (merged["MOTHER_PRE"] == 0)
            & (merged["MC_DOB"] < collection_start)
        ]
        err_list = err_co["orig_idx"].unique().tolist()
        err_list.sort()
        return {"Header": err_list}


def test_validate():
    import pandas as pd

    header = pd.DataFrame(
        [
            {"CHILD": "111", "MOTHER": 1, "MC_DOB": pd.NA},  # 0
            {"CHILD": "222", "MOTHER": "1", "MC_DOB": "04/01/2020"},  # 1 Fail
            {"CHILD": "333", "MOTHER": 0, "MC_DOB": pd.NA},  # 2
            {"CHILD": "444", "MOTHER": 1.0, "MC_DOB": "01/04/2020"},  # 3
        ]
    )
    header_last = pd.DataFrame(
        [
            {"CHILD": "111", "MOTHER": "1"},  # 0
            {"CHILD": "222", "MOTHER": 0.0},  # 1
            {"CHILD": "333", "MOTHER": "1"},  # 2
            {"CHILD": "444", "MOTHER": "0"},  # 3
        ]
    )
    metadata = {"collection_start": "01/04/2020"}

    fake_dfs = {"Header": header, "Header_last": header_last, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Header": [1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="204",
    message="Ethnic origin code disagrees with the ethnic origin already recorded for this child.",
    affected_fields=["ETHNIC"],
    tables=["Header", "HEader_last"],
)
def validate(dfs):
    if "Header" not in dfs or "Header_last" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        header_last = dfs["Header_last"]

        header_merged = (
            header.reset_index()
            .merge(
                header_last,
                how="left",
                on=["CHILD"],
                suffixes=("", "_last"),
                indicator=True,
            )
            .set_index("index")
        )

        in_both_years = header_merged["_merge"] == "both"
        ethnic_is_different = (
            header_merged["ETHNIC"].astype(str).str.upper()
            != header_merged["ETHNIC_last"].astype(str).str.upper()
        )

        error_mask = in_both_years & ethnic_is_different

        error_locations = header.index[error_mask]

        return {"Header": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "106", "108", "109", "110"],
            "ETHNIC": [
                "WBRI",
                "WBRI",
                "nobt",
                "AINS",
                pd.NA,
                "BOTH",
                pd.NA,
                "BCRB",
                "MWBC",
            ],
        }
    )

    fake_data_prev = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "107", "108", "109", "110"],
            "ETHNIC": [
                "WBRI",
                "NOBT",
                "NOBT",
                "AINS",
                pd.NA,
                "REFU",
                "MOTH",
                pd.NA,
                "MWBA",
            ],
        }
    )

    fake_dfs = {"Header": fake_data, "Header_last": fake_data_prev}

    result = validate(fake_dfs)

    assert result == {"Header": [1, 6, 7, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="526",
    message="Child is missing a placement provider code for at least one episode.",
    affected_fields=["PLACE", "PLACE_PROVIDER"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        error_mask = (
            ~epi["PLACE"].isin(["T0", "T1", "T2", "T3", "T4", "Z1"])
            & epi["PLACE_PROVIDER"].isna()
        )
        return {"Episodes": epi.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE": ["E1", "P0", "T1", "T3", "P1", "Z1", "P0"],
            "PLACE_PROVIDER": ["PR1", "PR2", pd.NA, "PR0", pd.NA, pd.NA, pd.NA],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    assert validate(fake_dfs) == {"Episodes": [4, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1009",
    message="Reason for placement change is not a valid code.",
    affected_fields=["REASON_PLACE_CHANGE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}

    episodes = dfs["Episodes"]
    code_list = [
        "CARPL",
        "CLOSE",
        "ALLEG",
        "STAND",
        "APPRR",
        "CREQB",
        "CREQO",
        "CHILD",
        "LAREQ",
        "PLACE",
        "CUSTOD",
        "OTHER",
    ]

    mask = (
        episodes["REASON_PLACE_CHANGE"].isin(code_list)
        | episodes["REASON_PLACE_CHANGE"].isna()
    )

    validation_error_mask = ~mask
    validation_error_locations = episodes.index[validation_error_mask]

    return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "REASON_PLACE_CHANGE": ["CARPL", "OTHER", "other", "NA", "", pd.NA],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [2, 3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="104",
    message="Date for Unaccompanied Asylum-Seeking Children (UASC) status ceased is not a valid date.",
    affected_fields=["DUC"],
    tables=["UASC"],
)
def validate(dfs):
    if "UASC" not in dfs:
        return {}
    else:
        uasc = dfs["UASC"]
        uasc["DUC_dt"] = pd.to_datetime(uasc["DUC"], format="%d/%m/%Y", errors="coerce")
        collection_start = pd.to_datetime(
            dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
        )
        mask = (uasc["DUC_dt"].isna() & uasc["DUC"].notna()) | (
            uasc["DUC_dt"] < collection_start
        )

        return {"UASC": uasc.index[mask].to_list()}


def test_validate():
    import pandas as pd

    fake_uasc = pd.DataFrame(
        {
            "DUC": [
                "01/03/2019",
                "19/02/2020",
                "03/04/2019",
                "01/01/2019",
                pd.NA,
                "INNVLAID/DATE/2077",
            ],
        }
    )

    metadata = {"collection_start": "01/04/2019"}

    fake_dfs = {"UASC": fake_uasc, "metadata": metadata}

    result = validate(fake_dfs)
    assert result == {"UASC": [0, 3, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="219",
    message="The Ofsted Unique reference number (URN) provided for the child's placement does not match the placement type recorded.",
    affected_fields=["URN", "PLACE"],
    tables=["Episodes", "Provider Info"],
)
def validate(dfs):
    if ("Episodes" not in dfs) or ("provider_info" not in dfs["metadata"]):
        return {}
    else:
        episodes = dfs["Episodes"]
        provider_info = dfs["metadata"]["provider_info"]

        # merge
        episodes["index_eps"] = episodes.index
        episodes = episodes[episodes["URN"].notna() & (episodes["URN"] != "XXXXXXX")]
        episodes = episodes.merge(provider_info, on="URN", how="left")
        # If <URN> provided and <URN> not = 'XXXXXXX' then <PL> must = any URN Lookup <PLACEMENT CODE> of matching URN Lookup <URN>
        place_valid = pd.Series(
            [
                False if (pd.isna(pl) or pd.isna(valid)) else pl in valid.split(",")
                for pl, valid in zip(episodes["PLACE"], episodes["PLACE_CODES"])
            ]
        )

        eps_error_locations = episodes.loc[~place_valid, "index_eps"]
        return {"Episodes": eps_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": "1111",
                "PLACE": "PR5",
                "URN": 1,
            },  # 0 fail
            {
                "CHILD": "1111",
                "PLACE": "PR5",
                "URN": pd.NA,
            },  # 1 ignore: URN
            {
                "CHILD": "1111",
                "PLACE": "PR7",
                "URN": 3,
            },  # 2 fail
            {
                "CHILD": "2222",
                "PLACE": "PR5",
                "URN": "XXXXXXX",
            },  # 3 ignore: URN
            {
                "CHILD": "3333",
                "PLACE": "PR1",
                "URN": 2,
            },  # 4 pass
            {
                "CHILD": "3333",
                "PLACE": "PR3",
                "URN": 2,
            },  # 5 pass
            {
                "CHILD": "4444",
                "PLACE": "PR5",
                "URN": 1,
            },  # 6 fail
            {
                "CHILD": "5555",
                "PLACE": "PR5",
                "URN": 4,
            },  # 7 fail - PLACE_CODES should not be null so probly needs a look
            {
                "CHILD": "5555",
                "PLACE": "PR2",
                "URN": 1,
            },  # 8 pass
        ]
    )
    fake_provider_info = pd.DataFrame(
        [
            {
                "URN": 1,
                "PLACE_CODES": "PR2",
            },  # 0
            {
                "URN": 2,
                "PLACE_CODES": "PR1,PR3,PR5",
            },  # 1
            {
                "URN": 3,
                "PLACE_CODES": "PR5,PR4,PR2",
            },  # 2
            {
                "URN": 4,
                "PLACE_CODES": pd.NA,
            },  # 3
        ]
    )
    metadata = {"provider_info": fake_provider_info}

    fake_dfs = {"Episodes": fake_data_eps, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 2, 6, 7]}



from dataclasses import dataclass
from functools import wraps
from typing import Callable, Optional


@dataclass(frozen=True, eq=True)
class RuleDefinition:
    """
    A dataclass type class used in each validation to assign information about
    each validation rule to the rule.

    :param str code: The rule code for each rule.
    :param Callable func: logic of the validation rule function.
    :param str message: The message to be displayed if rule is flagged.
    :param str affected_fields: The fields/columns affected by a validation rule.

    :returns: RuleDefinition object containing information about validation rules.
    :rtype: dataclass object.
    """

    code: str
    func: Callable
    message: Optional[str] = None
    affected_fields: Optional[list[str]] = None
    tables: Optional[list[str]] = (None,)


def rule_definition(
    code: str,
    message: Optional[str] = None,
    affected_fields: Optional[list[str]] = None,
    tables: Optional[list[str]] = None,
):
    """
    Creates the rule definition for validation rules using RuleDefinition class as a template.

    :param str code: The rule code for each rule.
    :param str message: The message displayed for each validation rule.
    :param str affected_fields: The fields/columns affected by a validation rule.

    :returns: RuleDefinition object containing information about validation rules.
    :rtype: RuleDefiniton class object.
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        definition = RuleDefinition(
            code=str(code),
            func=func,
            message=message,
            affected_fields=affected_fields,
            tables=tables,
        )
        # when validator funcs are created, give them a unique attribute that they can be
        # recognised by when the file is read later.
        wrapper.rule = definition
        return wrapper

    return decorator


@dataclass(eq=True)
class YearConfig:
    deleted: list[str]
    added_or_modified: dict[str, RuleDefinition]



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="215",
    message="Child has care leaver information but one or more data items relating to children looked after for 12 months have been completed.",
    affected_fields=[
        "IN_TOUCH",
        "ACTIV",
        "ACCOM",
        "CONVICTED",
        "HEALTH_CHECK",
        "IMMUNISATIONS",
        "TEETH_CHECK",
        "HEALTH_ASSESSMENT",
        "SUBSTANCE_MISUSE",
        "INTERVENTION_RECEIVED",
        "INTERVENTION_OFFERED",
    ],
    tables=["OC2", "OC3"],
)
def validate(dfs):
    if "OC3" not in dfs or "OC2" not in dfs:
        return {}
    else:
        oc3 = dfs["OC3"]
        oc2 = dfs["OC2"]
        # prepare to merge
        oc3.reset_index(inplace=True)
        oc2.reset_index(inplace=True)
        merged = oc3.merge(oc2, on="CHILD", how="left", suffixes=["_3", "_2"])
        # If any of <IN_TOUCH>, <ACTIV> or <ACCOM> have been provided then <CONVICTED>; <HEALTH_CHECK>; <IMMUNISATIONS>; <TEETH_CHECK>; <HEALTH_ASSESSMENT>; <SUBSTANCE MISUSE>; <INTERVENTION_RECEIVED>; <INTERVENTION_OFFERED>; should not be provided
        mask = (
            merged["IN_TOUCH"].notna()
            | merged["ACTIV"].notna()
            | merged["ACCOM"].notna()
        ) & (
            merged["CONVICTED"].notna()
            | merged["HEALTH_CHECK"].notna()
            | merged["IMMUNISATIONS"].notna()
            | merged["TEETH_CHECK"].notna()
            | merged["HEALTH_ASSESSMENT"].notna()
            | merged["SUBSTANCE_MISUSE"].notna()
            | merged["INTERVENTION_RECEIVED"].notna()
            | merged["INTERVENTION_OFFERED"].notna()
        )

        # error locations
        oc3_error_locs = merged.loc[mask, "index_3"]
        oc2_error_locs = merged.loc[mask, "index_2"]
        return {"OC3": oc3_error_locs.tolist(), "OC2": oc2_error_locs.tolist()}


def test_validate():
    import pandas as pd

    fake_data_oc2 = pd.DataFrame(
        {
            "CHILD": ["A", "B", "C", "D", "E"],
            "IMMUNISATIONS": ["1", pd.NA, "1", pd.NA, "1"],
            "TEETH_CHECK": ["1", pd.NA, "1", "1", "1"],
            "HEALTH_ASSESSMENT": ["1", pd.NA, "1", pd.NA, "1"],
            "SUBSTANCE_MISUSE": [pd.NA, pd.NA, "1", "1", "1"],
            "CONVICTED": [pd.NA, pd.NA, "1", "1", pd.NA],
            "HEALTH_CHECK": [pd.NA, pd.NA, "1", "1", pd.NA],
            "INTERVENTION_RECEIVED": ["1", pd.NA, "1", "1", pd.NA],
            "INTERVENTION_OFFERED": [pd.NA, "1", "1", "1", pd.NA],
        }
    )
    fake_data_oc3 = pd.DataFrame(
        {
            "CHILD": ["A", "B", "C", "D", "E"],
            "IN_TOUCH": ["DIED", "Yes", "RHOM", pd.NA, pd.NA],
            "ACTIV": [pd.NA, pd.NA, "XXX", pd.NA, pd.NA],
            "ACCOM": [pd.NA, pd.NA, pd.NA, "XXX", pd.NA],
        }
    )
    fake_dfs = {"OC3": fake_data_oc3, "OC2": fake_data_oc2}

    result = validate(fake_dfs)
    assert result == {"OC3": [0, 1, 2, 3], "OC2": [0, 1, 2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="558",
    message="If a child has been adopted, then the decision to place them for adoption has not been disrupted and the date of the decision that a child should no longer be placed for adoption should be left blank.",
    affected_fields=["DATE_PLACED_CEASED", "REC"],
    tables=["Episodes", "PlacedAdoption"],
)
def validate(dfs):
    # If latest episode <REC> = ‘E11’ or ‘E12’ then if number of <DATE_PLACED_CEASED> in the year is > 1 then latest
    # <DATE_PLACED_CEASED> must be   NULL
    if "Episodes" not in dfs or "PlacedAdoption" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        placedAdoptions = dfs["PlacedAdoption"]

        episodes = episodes.reset_index()

        rec_codes = ["E11", "E12"]

        placeEpisodes = episodes[episodes["REC"].isin(rec_codes)]

        merged = placeEpisodes.merge(placedAdoptions, how="left", on="CHILD").set_index(
            "index"
        )

        # List of children who have null dates to be excluded (assumes the null is their most recent)
        children_with_null = merged[(merged["DATE_PLACED_CEASED"].isna())][
            "CHILD"
        ].tolist()

        episodes_not_null = merged[~(merged["CHILD"].isin(children_with_null))]

        not_null = episodes_not_null.copy()

        not_null["COUNT"] = episodes_not_null.groupby(["CHILD"], group_keys=False)[
            "DATE_PLACED_CEASED"
        ].transform("count")

        episodes_with_errors = not_null[not_null["COUNT"] > 1]

        error_mask = episodes.index.isin(episodes_with_errors.index)

        error_locations = episodes.index[error_mask]

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        {
            "CHILD": [
                "0",
                "A",
                "B",
                "C",
                "D",
                "E",
                "F",
                "G",
                "H",
                "I",
            ],
            "REC": [
                "x",
                "E11",
                "E12",
                "E11",
                "E12",
                "E11",
                "E12",
                "E11",
                "E11",
                "E12",
            ],
        }
    )
    fake_data_placed_adoption = pd.DataFrame(
        {
            "CHILD": [
                "A",
                "B",
                "C",
                "D",
                "E",
                "F",
                "G",
                "H",
                "D",
                "E",
                "I",
                "I",
                "I",
                "I",
            ],
            "DATE_PLACED_CEASED": [
                pd.NA,
                pd.NA,
                pd.NA,
                "01/01/2020",
                "01/01/2020",
                "01/01/2020",
                "15/04/2020",
                pd.NA,
                "28th Jan 1930",
                "02/01/1930",
                "01/01/1930",
                "01/01/1930",
                "02/01/1930",
                pd.NA,
            ],
        }
    )

    fake_dfs = {
        "Episodes": fake_data_episodes,
        "PlacedAdoption": fake_data_placed_adoption,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [4, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="602",
    message="The episode data submitted for this child does not show that he/she was adopted during the year.",
    affected_fields=["CHILD"],
    tables=["Episodes", "AD1"],
)
def validate(dfs):
    if "Episodes" not in dfs or "AD1" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        ad1 = dfs["AD1"]
        epi["DEC"] = pd.to_datetime(epi["DEC"], format="%d/%m/%Y", errors="coerce")
        collection_start = pd.to_datetime(
            dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
        )
        collection_end = pd.to_datetime(
            dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
        )

        mask1 = (epi["DEC"] <= collection_end) & (epi["DEC"] >= collection_start)
        mask2 = epi["REC"].isin(["E11", "E12"])
        adoption_eps = epi[mask1 & mask2]

        adoption_fields = [
            "DATE_INT",
            "DATE_MATCH",
            "FOSTER_CARE",
            "NB_ADOPTR",
            "SEX_ADOPTR",
            "LS_ADOPTR",
        ]

        err_list = (
            ad1.reset_index()
            .merge(adoption_eps, how="left", on="CHILD", indicator=True)
            .query("_merge == 'left_only'")
            .dropna(subset=adoption_fields, how="all")
            .set_index("index")
            .index.unique()
            .to_list()
        )

        return {"AD1": err_list}


def test_validate():
    import pandas as pd

    fake_epi = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "444", "555", "666", "888"],
            "DEC": [
                "08/06/2020",
                "09/07/2020",
                "09/08/2020",
                "25/03/2021",
                "22/03/2020",
                "25/04/2021",
                pd.NA,
            ],
            "REC": ["E11", "oo", "T4", "E11", "oo", pd.NA, "E11"],
        }
    )
    fake_ad1 = pd.DataFrame(
        {
            "CHILD": ["111", "123", "222", "333", "345", "444", "555", "666", "777"],
            "DATE_INT": [pd.NA, "xx", pd.NA, pd.NA, "xx", "oo", pd.NA, "xx", "xx"],
        }
    )

    other_than_DATE_INT = [
        "DATE_MATCH",
        "FOSTER_CARE",
        "NB_ADOPTR",
        "SEX_ADOPTR",
        "LS_ADOPTR",
    ]

    for c in other_than_DATE_INT:
        fake_ad1[c] = pd.NA

    metadata = {"collection_start": "01/04/2020", "collection_end": "31/03/2021"}

    fake_dfs = {"Episodes": fake_epi, "AD1": fake_ad1, "metadata": metadata}

    assert validate(fake_dfs) == {"AD1": [1, 4, 7, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="353",
    message="No episode submitted can start before 14 October 1991.",
    affected_fields=["DECOM"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        min_decom_allowed = pd.to_datetime(
            "14/10/1991", format="%d/%m/%Y", errors="coerce"
        )
        error_mask = epi["DECOM"] < min_decom_allowed
        return {"Episodes": epi.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"DECOM": pd.NA},  # 0
            {"DECOM": "02/06/1980"},  # 1   Fails
            {"DECOM": "06/06/1890"},  # 2   Fails
            {"DECOM": "08/06/2020"},  # 3
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.types import MissingMetadataError


@rule_definition(
    code="105",
    message="Data entry for Unaccompanied Asylum-Seeking Children (UASC) status of child is invalid or has not been completed.",
    affected_fields=["UASC"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    else:
        try:
            file_format = dfs["metadata"]["file_format"]
        except KeyError as e:
            raise MissingMetadataError(*e.args)
        if file_format == "csv":
            return {}

        header = dfs["Header"]
        code_list = [0, 1]

        mask = ~pd.to_numeric(header["UASC"], errors="coerce").isin(code_list)
        error_locs = header.index[mask]

        return {"Header": error_locs.tolist()}


def test_validate():
    import pandas as pd

    fake_header = pd.DataFrame({"UASC": [0, 1, pd.NA, "", "0", "1", "2", 2]})

    fake_dfs = {"Header": fake_header, "metadata": {"file_format": "csv"}}
    result = validate(fake_dfs)
    assert result == {}

    fake_dfs = {"Header": fake_header, "metadata": {"file_format": "xml"}}
    result = validate(fake_dfs)
    assert result == {"Header": [2, 3, 6, 7]}



from lac_validator.fixtures import fake_INT_file, fake_INT_header
from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="INT06",
    message="Internal Check: Child in OC3 does not exist in Header.",
    affected_fields=["CHILD"],
    tables=["Header", "OC3"],
)
def validate(dfs):
    if "Header" not in dfs or "OC3" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        file = dfs["OC3"]

        file["index_file"] = file.index

        merged = header.merge(
            file[["CHILD", "index_file"]],
            on="CHILD",
            indicator=True,
            how="right",
            suffixes=["_header", "_file"],
        )

        mask = merged["_merge"] == "right_only"
        eps_error_locations = merged.loc[mask, "index_file"]
        return {"OC3": eps_error_locations.unique().tolist()}


def test_validate():
    fake_dfs = {"Header": fake_INT_header.copy(), "OC3": fake_INT_file.copy()}
    result = validate(fake_dfs)
    assert result == {"OC3": [3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.utils import (
    add_col_to_tables_CONTINUOUSLY_LOOKED_AFTER as add_CLA_column,  # Check 'Episodes' present before use!
)


@rule_definition(
    code="543",
    message="Child is aged 10 or over at 31 March and has been looked after continuously for 12 months yet conviction information has not been completed.",
    affected_fields=["DOB", "CONVICTED"],
    tables=["Episodes", "OC2"],
)
def validate(dfs):
    if "OC2" not in dfs or "Episodes" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]
        collection_end = dfs["metadata"]["collection_end"]
        # add CLA column
        oc2 = add_CLA_column(dfs, "OC2")

        # to datetime
        oc2["DOB"] = pd.to_datetime(oc2["DOB"], format="%d/%m/%Y", errors="coerce")
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )

        # If <DOB> >= 10 years prior to <COLLECTION_END_DATE>and<CONTINUOUSLY_LOOKED_AFTER> = 'Y' then <CONVICTED> should be provided
        mask = (
            (collection_end > (oc2["DOB"] + pd.offsets.DateOffset(years=10)))
            & oc2["CONTINUOUSLY_LOOKED_AFTER"]
            & oc2["CONVICTED"].isna()
        )
        error_locations = oc2.index[mask]
        return {"OC2": error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DECOM": "01/03/1980",
                "DEC": "31/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },  # CLA
            {
                "CHILD": 102,
                "DECOM": "01/03/1980",
                "DEC": "30/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },  # CLA
            {
                "CHILD": 103,
                "DECOM": "01/03/1980",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },  # CLA
            {
                "CHILD": 104,
                "DECOM": "01/02/1970",
                "DEC": pd.NA,
                "LS": "o",
                "REC": "!!",
                "RNE": "o",
            },  # CLA
            {
                "CHILD": 105,
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },  # CLA
            {
                "CHILD": 105,
                "DECOM": "01/01/1981",
                "DEC": "01/01/1983",
                "LS": "o",
                "REC": "oo",
                "RNE": "o",
            },  # CLA
            {
                "CHILD": 106,
                "DECOM": "01/03/1980",
                "DEC": "01/01/1982",
                "LS": "V3",
                "REC": "X1",
                "RNE": "o",
            },  # not CLA: V3
            {
                "CHILD": 107,
                "DECOM": "01/03/1980",
                "DEC": "01/01/1982",
                "LS": "V3",
                "REC": "!!",
                "RNE": "o",
            },  # not CLA: REC
        ]
    )
    fake_data = pd.DataFrame(
        {
            "CHILD": [101, 102, 103, 104, 105, 106, 107, 333],
            "DOB": [
                "08/03/1973",
                "22/06/1968",
                pd.NA,
                "13/10/1968",
                "10/01/1966",
                "01/01/1965",
                "01/01/1965",
                "01/01/1965",
            ],
            "CONVICTED": [pd.NA, pd.NA, pd.NA, 1, pd.NA, pd.NA, pd.NA, pd.NA],
            # 0 pass: under 10
            # 1 fail : CLA is true, over 10, and CONVICTED is not provided
            # 2 pass: DOB is nan
            # 3 pass: under 5 (born in future), but CONVICTED provided
            # 4 fail: CLA is true, over 10, and CONVICTED is not provided
            # 5 pass: not CLA
            # 6 pass: not CLA
            # 7 pass: not in episodes -> not CLA
        }
    )

    metadata = {"collection_start": "01/04/1980", "collection_end": "31/03/1981"}

    fake_dfs = {"OC2": fake_data, "metadata": metadata, "Episodes": fake_data_episodes}

    result = validate(fake_dfs)

    assert result == {"OC2": [1, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.utils import (
    add_col_to_tables_CONTINUOUSLY_LOOKED_AFTER as add_CLA_column,  # Check 'Episodes' present before use!
)


@rule_definition(
    code="198",
    message="Child has not been looked after continuously for at least 12 months at 31 March but a reason "
    "for no Strengths and Difficulties (SDQ) score has been completed. ",
    affected_fields=["SDQ_REASON"],
    tables=["Episodes", "OC2"],
)
def validate(dfs):
    if "Episodes" not in dfs or "OC2" not in dfs:
        return {}

    oc2 = add_CLA_column(dfs, "OC2")

    error_mask = oc2["SDQ_REASON"].notna() & ~oc2["CONTINUOUSLY_LOOKED_AFTER"]

    error_locs = oc2.index[error_mask].to_list()

    return {"OC2": error_locs}


def test_validate():
    import pandas as pd

    metadata = {"collection_start": "01/04/1980", "collection_end": "31/03/1981"}

    eps = pd.DataFrame(
        [
            {
                "CHILD": "1",
                "DECOM": "01/03/1980",
                "DEC": "31/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "2",
                "DECOM": "01/03/1980",
                "DEC": "30/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "3",
                "DECOM": "01/03/1980",
                "DEC": "01/01/1981",
                "LS": "V3",
                "REC": "X1",
                "RNE": "o",
            },  # False
            {
                "CHILD": "4",
                "DECOM": "01/02/1970",
                "DEC": pd.NA,
                "LS": "o",
                "REC": "!!",
                "RNE": "o",
            },
            {
                "CHILD": "5555",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "5555",
                "DECOM": "01/01/1981",
                "DEC": pd.NA,
                "LS": "o",
                "REC": pd.NA,
                "RNE": "o",
            },
            {
                "CHILD": "6",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "!!",
                "RNE": "o",
            },  # !! - False
            {
                "CHILD": "6",
                "DECOM": "01/01/1981",
                "DEC": pd.NA,
                "LS": "o",
                "REC": pd.NA,
                "RNE": "o",
            },  # False
            {
                "CHILD": "7777",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },  # False
            {
                "CHILD": "7777",
                "DECOM": "01/01/1981",
                "DEC": "01/07/1981",
                "LS": "o",
                "REC": "o",
                "RNE": "S",
            },  # !! - False
            {
                "CHILD": "8",
                "DECOM": "01/01/1981",
                "DEC": "31/03/1999",
                "LS": "o",
                "REC": "o",
                "RNE": "S",
            },  # False
        ]
    )

    oc2 = pd.DataFrame(
        {
            "CHILD": ["999999", "1", "2", "3", "9999", "8"],
            "SDQ_REASON": ["!!!!!", pd.NA, "OO", "!!", pd.NA, "!!"],
        }
    )

    fake_dfs = {"Episodes": eps, "OC2": oc2, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"OC2": [0, 3, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="355",
    message="Episode appears to have lasted for less than 24 hours",
    affected_fields=["DECOM", "DEC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        df = dfs["Episodes"]
        mask = df["DECOM"].astype(str) == df["DEC"].astype(str)
        return {"Episodes": df.index[mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "DECOM": [
                "01/01/2000",
                "01/02/2000",
                "01/03/2000",
                "01/04/2000",
                "01/05/2000",
                "01/06/2000",
                "04/05/2000",
            ],
            "DEC": [
                "01/01/2000",
                "01/03/2000",
                pd.NA,
                "01/04/2000",
                "03/05/2000",
                "01/06/2000",
                "01/05/2000",
            ],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    assert validate(fake_dfs) == {"Episodes": [0, 3, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="181",
    message="Data items relating to children looked after continuously for 12 months should be completed with a 0 or 1.",
    affected_fields=[
        "CONVICTED",
        "HEALTH_CHECK",
        "IMMUNISATIONS",
        "TEETH_CHECK",
        "HEALTH_ASSESSMENT",
        "SUBSTANCE_MISUSE",
        "INTERVENTION_RECEIVED",
        "INTERVENTION_OFFERED",
    ],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]
        code_list = ["0", "1"]

        fields_of_interest = [
            "CONVICTED",
            "HEALTH_CHECK",
            "IMMUNISATIONS",
            "TEETH_CHECK",
            "HEALTH_ASSESSMENT",
            "SUBSTANCE_MISUSE",
            "INTERVENTION_RECEIVED",
            "INTERVENTION_OFFERED",
        ]

        error_mask = (
            oc2[fields_of_interest].notna()
            & ~oc2[fields_of_interest].astype(str).isin(["0", "1"])
        ).any(axis=1)

        validation_error_locations = oc2.index[error_mask]

        return {"OC2": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CONVICTED": [1, pd.NA, pd.NA, "2", "2"],
            "HEALTH_CHECK": [1, pd.NA, pd.NA, 2, "2"],
            "IMMUNISATIONS": [1, pd.NA, pd.NA, pd.NA, "2"],
            "TEETH_CHECK": ["1", pd.NA, "0", pd.NA, "2"],
            "HEALTH_ASSESSMENT": ["1", pd.NA, 1, pd.NA, "2"],
            "SUBSTANCE_MISUSE": [0, pd.NA, 0, "1", "2"],
            "INTERVENTION_RECEIVED": [0, pd.NA, "1", 1, "2"],
            "INTERVENTION_OFFERED": ["1", pd.NA, pd.NA, 0, "2"],
        }
    )

    fake_dfs = {"OC2": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC2": [3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.utils import add_col_to_episodes_CONTINUOUSLY_LOOKED_AFTER


@rule_definition(
    code="SW12STG2",
    message="If a child is looked after on 31 March then all social worker episodes for the full previous year should be reported, even those whilst they were not looked after",
    affected_fields=["SW_DECOM"],
    tables=["SWEpisodes", "Episodes"],
)
def validate(dfs):
    if ("SWEpisodes" not in dfs) | ("Episodes" not in dfs):
        return {}
    else:
        SWE = dfs["SWEpisodes"]
        epi = dfs["Episodes"]

        SWE["SW_DECOM"] = pd.to_datetime(
            SWE["SW_DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        epi["DEC"] = pd.to_datetime(epi["DEC"], format="%d/%m/%Y", errors="coerce")
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")

        epi = epi.reset_index()

        epi = epi[
            (~epi["LS"].isin(["V3", "V4"]))
            & ((epi["RNE"] == "S") | (epi["REC"] != "X1"))
        ]

        df = pd.merge(epi, SWE, how="left", on=["CHILD"])

        # finding episodes that start or end care and don't have a corresponding SWE episode
        swdecom_before_dec_nx1 = ((df["SW_DECOM"] <= df["DEC"])) & (df["REC"] != "X1")
        swdecom_after_decom_s = ((df["SW_DECOM"] >= df["DECOM"])) & (df["RNE"] == "S")

        error_rows = df[~swdecom_before_dec_nx1 & ~swdecom_after_decom_s]["index"]

        return {"Episodes": error_rows.tolist()}


def test_validate():
    import pandas as pd

    epi = pd.DataFrame(
        [
            {
                "CHILD": 1,
                "DECOM": "01/01/2024",
                "DEC": pd.NA,
                "REC": pd.NA,
                "RNE": "S",
                "LS": "C1",
            },  # Pass, has SW EP
            {
                "CHILD": 2,
                "DECOM": "01/01/2024",
                "DEC": pd.NA,
                "REC": pd.NA,
                "RNE": "S",
                "LS": "C1",
            },  # FAIL, no SW EP
            {
                "CHILD": 3,
                "DECOM": "01/01/2024",
                "DEC": pd.NA,
                "REC": pd.NA,
                "RNE": "S",
                "LS": "V4",
            },  # Pass, no SW EP but V4 (exempt)
            {
                "CHILD": 4,
                "DECOM": "01/01/2024",
                "DEC": "10/01/2024",
                "REC": "X1",
                "RNE": "P",
                "LS": "C1",
            },  # Pass, no SW EP but not a start or end placement
            {
                "CHILD": 4,
                "DECOM": "10/01/2024",
                "DEC": pd.NA,
                "REC": pd.NA,
                "RNE": "P",
                "LS": "C1",
            },  # Pass, no SW EP but not a start or end placement
            {
                "CHILD": 5,
                "DECOM": "01/01/2024",
                "DEC": "10/01/2024",
                "REC": "X1",
                "RNE": "S",
                "LS": "C1",
            },  # FAIL no SW EP (started after end)
            {
                "CHILD": 5,
                "DECOM": "10/01/2024",
                "DEC": "01/03/2024",
                "REC": "E4a",
                "RNE": "P",
                "LS": "C1",
            },  # Pass has SW EP
            {
                "CHILD": 6,
                "DECOM": "01/01/2024",
                "DEC": "10/01/2024",
                "REC": "X1",
                "RNE": "S",
                "LS": "C1",
            },  # Pass has SW EP
            {
                "CHILD": 6,
                "DECOM": "10/01/2024",
                "DEC": "01/03/2024",
                "REC": "E4a",
                "RNE": "P",
                "LS": "C1",
            },  # Pass has SW EP
            {
                "CHILD": 7,
                "DECOM": "01/04/2024",
                "DEC": pd.NA,
                "REC": "X1",
                "RNE": "S",
                "LS": "C1",
            },  # Edge case? FAIL but technically think should be fine if the SW_DEC (or no DEC) was after EPI DECOM (which we're not checking here)
        ]
    )

    sw_eps = pd.DataFrame(
        [
            {"CHILD": 1, "SW_DECOM": "01/01/2024"},
            {"CHILD": 5, "SW_DECOM": "01/02/2024"},
            {"CHILD": 6, "SW_DECOM": "01/02/2024"},
            {"CHILD": 7, "SW_DECOM": "01/02/2024"},
        ]
    )

    fake_dfs = {"SWEpisodes": sw_eps, "Episodes": epi}

    result = validate(fake_dfs)

    print(result)

    assert result == {"Episodes": [1, 4, 9]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW11bSTG2",
    message="Social worker episode has ended before 31 March and no subsequent social worker episode has been provided.",
    affected_fields=["SW_DEC"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]

        df["SW_DEC"] = pd.to_datetime(df["SW_DEC"], format="%d/%m/%Y", errors="coerce")
        df["SW_DECOM"] = pd.to_datetime(
            df["SW_DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        collection_end = dfs["metadata"]["collection_end"]
        collection_end = pd.to_datetime(collection_end, format="%d/%m/%Y")

        df["index"] = df.index

        # Children with an NA following a SW_DEC
        df = df.sort_values(["CHILD", "SW_DECOM"])

        df_lead = df.shift(1)

        df_lead = df_lead.reset_index()

        m_df = df.merge(
            df_lead,
            left_on=["index", "CHILD"],
            right_on=["level_0", "CHILD"],
            how="outer",
            suffixes=("", "_prev"),
        )

        m_df.dropna(subset="index", axis="rows", inplace=True)
        m_df["index"] = m_df["index"].astype("int")
        before_end = m_df[
            (m_df["SW_DEC_prev"] >= collection_end - pd.DateOffset(days=5))
            & (m_df["SW_DEC_prev"] <= collection_end)
        ]

        error_rows = before_end[before_end["SW_DECOM"].isna()]["index"]
        error_rows = error_rows.to_list()

        # need children who appear once in SW DEC after timeframe
        appearance_counts = df.value_counts("CHILD").reset_index()
        one_appearance = appearance_counts[appearance_counts[0] == 1]["CHILD"].tolist()
        children_appearing_once = df[
            df["CHILD"].isin(one_appearance)
            & (
                (df["SW_DEC"] >= collection_end - pd.DateOffset(days=5))
                & (df["SW_DEC"] <= collection_end)
            )
        ]["index"].tolist()

        # children who have multiple SW_DEC in the time period and no following SW_DECOM
        # sorting in reverse lets us find rows where there's multiple filled DEC/DECOM
        # in the timeframe with no DECOM following the last DEC by putting the
        df = df.sort_values(["CHILD", "SW_DECOM"], ascending=False)

        df_reverse_lead = df.shift(1)

        df_reverse_lead = df_reverse_lead.reset_index()

        m_reverse_df = df.merge(
            df_reverse_lead,
            left_on=["index", "CHILD"],
            right_on=["level_0", "CHILD"],
            how="outer",
            suffixes=("", "_prev"),
        )
        no_following_decom = m_reverse_df[
            (m_reverse_df["SW_DECOM"] < m_reverse_df["SW_DEC_prev"])
            & (
                (m_reverse_df["SW_DEC_prev"] >= collection_end - pd.DateOffset(days=5))
                & (m_reverse_df["SW_DEC_prev"] <= collection_end)
            )
        ]

        multiple_dec_in_timeframe = (
            no_following_decom["index_prev"].astype(int).to_list()
        )

        error_rows.extend(children_appearing_once)
        error_rows.extend(multiple_dec_in_timeframe)
        return {"SWEpisodes": error_rows}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"CHILD": "1", "SW_DECOM": "01/04/2000", "SW_DEC": "26/03/2000"},
            {
                "CHILD": "1",
                "SW_DECOM": "01/05/2000",
                "SW_DEC": pd.NA,
            },  # Passes, rule doesn't suggest a timeframe needed for SW_DECOM
            {"CHILD": "2", "SW_DECOM": "01/04/2000", "SW_DEC": "25/03/2000"},
            {"CHILD": "2", "SW_DECOM": "02/04/2000", "SW_DEC": "03/04/2000"},
            {"CHILD": "3", "SW_DECOM": "01/04/2000", "SW_DEC": "1/04/2000"},
            {"CHILD": "3", "SW_DECOM": "02/04/2000", "SW_DEC": "03/04/2000"},
            {"CHILD": "4", "SW_DECOM": "01/04/2000", "SW_DEC": "30/03/2000"},
            {"CHILD": "4", "SW_DECOM": "30/03/2000", "SW_DEC": "03/04/2000"},
            {"CHILD": "5", "SW_DECOM": "01/04/2000", "SW_DEC": "30/03/2000"},
            {"CHILD": "6", "SW_DECOM": "01/04/2000", "SW_DEC": "26/03/2000"},
            {"CHILD": "6", "SW_DECOM": pd.NA, "SW_DEC": pd.NA},  # Fails, no SW_DECOM
            {"CHILD": "7", "SW_DECOM": "01/04/2000", "SW_DEC": "25/03/2000"},
            {
                "CHILD": "7",
                "SW_DECOM": pd.NA,
                "SW_DEC": pd.NA,
            },  # Passes, no SW_DEC outside of timeframe
            {"CHILD": "8", "SW_DECOM": "01/04/2000", "SW_DEC": "26/03/2000"},
            {"CHILD": "8", "SW_DECOM": "27/03/2000", "SW_DEC": "28/03/2000"},
            {
                "CHILD": "8",
                "SW_DECOM": pd.NA,
                "SW_DEC": pd.NA,
            },  # Fails, no SW_DECOM after DEC in timeframe
            {"CHILD": "9", "SW_DECOM": "01/03/2000", "SW_DEC": "25/03/2000"},
            {
                "CHILD": "9",
                "SW_DECOM": "26/03/2000",
                "SW_DEC": "27/03/2000",
            },  # Should fail, SW_DEC in timeframe no following DECOM
        ]
    )

    fake_meta = {"collection_end": "31/3/2000"}

    fake_dfs = {"SWEpisodes": fake_data, "metadata": fake_meta}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [10, 15, 8, 17]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="164",
    message="Distance is not valid. Please check a valid postcode has been entered. [NOTE: This check will result in false positives for children formerly UASC not identified as such in loaded data]",
    affected_fields=["PL_DISTANCE"],
    tables=["Episodes"],
)

# If not (<UASC> = '1' or <COLLECTION_YEAR> -1 <UASC> = '1' or <COLLECTION_YEAR> -2 <UASC> = '1') and if <LS> not = 'V3' or 'V4' then <PL_DISTANCE> must be between '0.0' and '999.9' Note: <PL_DISTANCE> is a calculated value based on Postcode.
def validate(dfs):
    # Header_last also used if present
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]

        # Note the initial positions. Freeze a copy of the index values into a column
        epi["orig_idx"] = epi.index

        # Work out who is UASC or formerly UASC
        header = pd.DataFrame()
        if "Header" in dfs:
            header_current = dfs["Header"]
            header = pd.concat((header, header_current), axis=0)
        elif "UASC" in dfs:
            uasc = dfs["UASC"]
            uasc = uasc.loc[
                uasc.drop("CHILD", axis="columns").notna().any(axis=1), ["CHILD"]
            ].copy()
            uasc.loc[:, "UASC"] = "1"
            header = pd.concat((header, uasc), axis=0)

        if "Header_last" in dfs:
            header = pd.concat((header, dfs["Header_last"]), axis=0)
        elif "UASC_last" in dfs:
            uasc = dfs["UASC_last"]
            uasc = uasc.loc[
                uasc.drop("CHILD", axis="columns").notna().any(axis=1), ["CHILD"]
            ].copy()
            uasc.loc[:, "UASC"] = "1"
            header = pd.concat((header, uasc), axis=0)

        if "UASC" in header.columns:
            header = header[header.UASC == "1"].drop_duplicates("CHILD")
            # drop all CHILD IDs that ever have UASC == 1
            epi = epi.merge(
                header[["CHILD"]], how="left", on="CHILD", indicator=True
            ).query('_merge == "left_only"')
        else:
            # if theres no UASC column in header, either from XML data, inferred for CSV in ingress, or added above
            # then we can't identify anyone as UASC/formerly UASC
            return {}
        # PL_DISTANCE is added when the uploaded files are read into the tool. The code that does this is found in datastore.py
        epi = epi[epi["PL_DISTANCE"].notna()]
        check_in_range = (epi["PL_DISTANCE"].astype("float") < 0.0) | (
            epi["PL_DISTANCE"].astype("float") > 999.9
        )
        err_list = epi.loc[(check_in_range), "orig_idx"].sort_values().unique().tolist()

        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_episodes = pd.DataFrame(
        [
            {"CHILD": "111", "LS": "V3", "PL_DISTANCE": "0.0"},  # 0 ignore: LS
            {"CHILD": "222", "LS": "V1", "PL_DISTANCE": pd.NA},  # 1
            {"CHILD": "222", "LS": "V5", "PL_DISTANCE": "-2"},  # 2 fail
            {"CHILD": "333", "LS": "V1", "PL_DISTANCE": "3.5"},  # 3 pass
            {"CHILD": "333", "LS": "V2", "PL_DISTANCE": "1000"},  # 4 fail
            {"CHILD": "345", "LS": "V7", "PL_DISTANCE": "1380"},  # 5 ignore: UASC
        ]
    )
    fake_header = pd.DataFrame(
        [
            {"CHILD": "111", "UASC": "0"},
            {"CHILD": "222", "UASC": "0"},
            {"CHILD": "333", "UASC": "0"},
            {"CHILD": "345", "UASC": "0"},
        ]
    )
    fake_header_last = pd.DataFrame(
        [
            {"CHILD": "111", "UASC": "0"},
            {"CHILD": "222", "UASC": "0"},
            {"CHILD": "333", "UASC": "0"},
            {"CHILD": "345", "UASC": "1"},
        ]
    )

    fake_dfs = {
        "Episodes": fake_episodes,
        "Header": fake_header,
        "Header_last": fake_header_last,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [2, 4]}

    uasc_last = pd.DataFrame(
        [
            {"CHILD": "333", "DUC": "something"},  # [3] in episodes - passes now
            {"CHILD": "345", "DUC": pd.NA},  # [8] in episodes - still fails
        ]
    )

    dfs = {"Episodes": fake_episodes, "Header": fake_header, "UASC_last": uasc_last}
    result = validate(dfs)
    assert result == {"Episodes": [2, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="231",
    message="We expect that unaccompanied asylum seeking children are generally older young people. Please check the date of birth is correct. If the date of birth is correct then please provide details in a bypass request.",
    affected_fields=["DOB", "UASC_STATUS"],
    tables=["Header"],
)
def validate(dfs):
    # If <UASC_STATUS> = 1 then <DOB> should be > 11 years prior to <COLLECTION_END_DATE>
    if "Header" not in dfs:
        return {}
    else:
        df = dfs["Header"]

        collection_end = pd.to_datetime(
            dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
        )

        df["DOB"] = pd.to_datetime(df["DOB"], format="%d/%m/%Y", errors="coerce")

        under_11_uasc = df[
            ((df["DOB"] > collection_end - pd.DateOffset(years=11)))
            & (df["UASC_STATUS"].astype(str) == "1")
        ].index

        return {"Header": under_11_uasc.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"DOB": "01/04/2014", "UASC_STATUS": "1"},  # under 11 uasc, fails
            {
                "DOB": "31/03/2014",
                "UASC_STATUS": "1",
            },  # 11 day of collection end, uasc, according to stric rule logic, passes
            {"DOB": "01/04/2014", "UASC_STATUS": "0"},
            {"DOB": "30/03/2014", "UASC_STATUS": "1"},
        ]
    )  # over 11, uasc, passes

    fake_metadata = {"collection_end": "31/03/2025"}

    fake_dfs = {"Header": fake_data, "metadata": fake_metadata}

    result = validate(fake_dfs)

    assert result == {"Header": [0]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="362",
    message="Emergency protection order (EPO) lasted longer than 21 days",
    affected_fields=["DECOM", "LS", "DEC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]

        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi["DEC"] = pd.to_datetime(epi["DEC"], format="%d/%m/%Y", errors="coerce")
        collection_end = pd.to_datetime(
            dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
        )

        epi.sort_values(["CHILD", "DECOM"], inplace=True)
        epi.reset_index(inplace=True)
        epi.reset_index(inplace=True)

        epi["LAG"] = epi["level_0"] - 1

        epi["DEC"].fillna(collection_end, inplace=True)

        err_co = epi.merge(
            epi,
            how="left",
            left_on="level_0",
            right_on="LAG",
            suffixes=["", "_NEXT"],
        ).query("LS == 'L2'")

        # Create a partition "FLOWS" for two or more separate flow sequences of L2 code dates within the same child.
        # when the dec / decom_next dates stop flowing or the child id changes
        # the cumsum is incremented this can then be used as the partition.
        err_co["FLOWS"] = (err_co["DEC"] == err_co["DECOM_NEXT"]) & (
            err_co["CHILD"] == err_co["CHILD_NEXT"]
        )
        err_co["FLOWS"] = err_co["FLOWS"].shift(1)
        err_co["FLOWS"].fillna(False, inplace=True)
        err_co["FLOWS"] = ~err_co["FLOWS"]
        err_co["FLOWS"] = err_co["FLOWS"].astype(int).cumsum()

        # Calc the min decom and max dec in each group so the days between them can be calculated.
        grp_decom = (
            err_co.groupby(["CHILD", "FLOWS"])["DECOM"]
            .min()
            .to_frame(name="MIN_DECOM")
            .reset_index()
        )
        grp_dec = (
            err_co.groupby(["CHILD", "FLOWS"])["DEC"]
            .max()
            .to_frame(name="MAX_DEC")
            .reset_index()
        )
        grp_len_l2 = grp_decom.merge(grp_dec, how="inner", on=["CHILD", "FLOWS"])

        # Throw out anything <= 21 days.
        grp_len_l2["DAY_DIF"] = (
            grp_len_l2["MAX_DEC"] - grp_len_l2["MIN_DECOM"]
        ).dt.days
        grp_len_l2 = grp_len_l2.query("DAY_DIF > 21").copy()

        # Inner join back to the err_co and get the original index out.
        err_co["MERGE_KEY"] = err_co["CHILD"].astype(str) + err_co["FLOWS"].astype(str)
        grp_len_l2["MERGE_KEY"] = grp_len_l2["CHILD"].astype(str) + grp_len_l2[
            "FLOWS"
        ].astype(str)
        err_final = err_co.merge(
            grp_len_l2, how="inner", on=["MERGE_KEY"], suffixes=["", "_IG"]
        )

        err_list = err_final["index"].unique().tolist()
        err_list.sort()

        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DECOM": "01/06/2020",
                "DEC": "04/06/2020",
                "LS": "L2",
            },  # 0  CHANGE |a
            {
                "CHILD": "111",
                "DECOM": "04/06/2020",
                "DEC": "05/06/2020",
                "LS": "L2",
            },  # 1         |a Passes as <=21
            {
                "CHILD": "111",
                "DECOM": "06/07/2020",
                "DEC": "07/07/2020",
                "LS": "E2",
            },  # 2
            {
                "CHILD": "111",
                "DECOM": "08/09/2020",
                "DEC": "09/10/2020",
                "LS": "L2",
            },  # 3  CHANGE |b  Fail
            {
                "CHILD": "111",
                "DECOM": "09/10/2020",
                "DEC": "12/11/2020",
                "LS": "L2",
            },  # 4         |b  Fail
            {
                "CHILD": "333",
                "DECOM": "06/06/2020",
                "DEC": "10/06/2020",
                "LS": "L2",
            },  # 5  CHANGE |c  Fail
            {
                "CHILD": "333",
                "DECOM": "10/06/2020",
                "DEC": "12/10/2020",
                "LS": "L2",
            },  # 6         |c  Fail
            {
                "CHILD": "444",
                "DECOM": "07/06/2020",
                "DEC": "08/07/2020",
                "LS": "L2",
            },  # 7  CHANGE |d  Fail
            {
                "CHILD": "444",
                "DECOM": "08/08/2020",
                "DEC": "09/08/2020",
                "LS": "E4",
            },  # 8
            {
                "CHILD": "444",
                "DECOM": "09/08/2020",
                "DEC": "15/08/2020",
                "LS": "L2",
            },  # 9  CHANGE |e  Fail
            {
                "CHILD": "444",
                "DECOM": "15/08/2020",
                "DEC": pd.NA,
                "LS": "L2",
            },  # 10        |e  Fail
            {
                "CHILD": "555",
                "DECOM": "15/03/2021",
                "DEC": pd.NA,
                "LS": "L2",
            },  # 11 CHANGE |f  Passes as <=21
        ]
    )
    metadata = {"collection_end": "31/03/2021"}

    fake_dfs = {"Episodes": fake_data, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Episodes": [3, 4, 5, 6, 7, 9, 10]}



import pandas as pd

from lac_validator.fixtures import current_episodes, previous_episodes
from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="503G",
    message="The distance in first episode does not match open episode at end of last year.",
    affected_fields=["PL_DISTANCE"],
    tables=["Episodes", "Episodes_last"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Episodes_last" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi_last = dfs["Episodes_last"]
        field = "PL_DISTANCE"

        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi_last["DECOM"] = pd.to_datetime(
            epi_last["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        epi_last["DEC"] = pd.to_datetime(
            epi_last["DEC"], format="%d/%m/%Y", errors="coerce"
        )

        epi.reset_index(inplace=True)

        first_ep_inds = epi.groupby(["CHILD"])["DECOM"].idxmin(skipna=True)
        min_decom = epi.loc[first_ep_inds, :]

        last_ep_inds = epi_last.groupby(["CHILD"])["DECOM"].idxmax(skipna=True)
        max_last_decom = epi_last.loc[last_ep_inds, :]

        merged_co = min_decom.merge(
            max_last_decom, how="inner", on=["CHILD"], suffixes=["", "_PRE"]
        )

        err_mask = (
            abs(
                merged_co["PL_DISTANCE"].astype(float, errors="ignore")
                - merged_co["PL_DISTANCE_PRE"].astype(float, errors="ignore")
            )
            >= 0.2
        ) | (
            merged_co["PL_DISTANCE_PRE"].isna() & merged_co["PL_DISTANCE"].notna()
            | (merged_co["PL_DISTANCE_PRE"].isna() & merged_co["PL_DISTANCE"].notna())
        )

        err_mask = err_mask & merged_co["DEC_PRE"].isna()

        err_list = merged_co["index"][err_mask].unique().tolist()
        err_list.sort()
        return {"Episodes": err_list}


def test_validate():
    fake_epi = current_episodes.copy()
    fake_epi_last = previous_episodes.copy()

    # For the test, this doesn't affect anything except how the data is handled
    # by the field_different function, so it doesn't matter that the years
    # in the test dfs are not accurate to this collection year
    # As seen in conftest collection year is read in as XXXX/XX and stored in metadata
    # as the first four characters using [:4] in _process_metadata in datastore```
    metadata = {"collectionYear": "2022"}

    fake_dfs = {
        "Episodes": fake_epi,
        "Episodes_last": fake_epi_last,
        "metadata": metadata,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [4, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="460",
    message="Reason episode ceased is that child stayed with current carers at age 18 (or above), but child is aged under 18.",
    affected_fields=["DEC", "REC"],
    tables=["Header", "Episodes"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    if "Episodes" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        episodes = dfs["Episodes"]

        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        header["DOB18"] = header["DOB"] + pd.DateOffset(years=18)

        episodes = episodes[episodes["REC"] == "E17"]

        episodes_merged = (
            episodes.reset_index()
            .merge(
                header,
                how="left",
                on=["CHILD"],
                suffixes=("", "_header"),
                indicator=True,
            )
            .set_index("index")
        )

        care_ended_under_18 = episodes_merged["DOB18"] > episodes_merged["DEC"]

        error_mask = care_ended_under_18

        error_locations = episodes.index[error_mask]

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "101", "102", "103", "104"],
            "REC": ["E17", "E17", "X1", pd.NA, "E17", "E17"],
            "DEC": [
                "16/03/2023",
                "17/06/2020",
                "20/03/2020",
                pd.NA,
                "23/08/2020",
                "23/08/2020",
            ],
        }
    )

    fake_data_child = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104"],
            "DOB": ["16/03/2005", "23/09/2002", "31/12/2000", pd.NA],
        }
    )

    fake_dfs = {"Episodes": fake_data, "Header": fake_data_child}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="392b",
    message="Child is looked after but no postcodes are recorded. [NOTE: This check may result in false "
    "positives for children formerly UASC, particularly if current & prior year UASC data not loaded]",
    affected_fields=["HOME_POST", "PL_POST"],
    tables=["Episodes"],
)
def validate(dfs):
    # Will also use Header/UASC and/or Header_last/UASC_last for former UASC status
    if "Episodes" not in dfs:
        return {}
    else:
        # If <LS> not = 'V3' or 'V4' and <UASC> = '0' and <COLLECTION YEAR> - 1 <UASC> = '0' and <COLLECTION YEAR> - 2 <UASC> = '0' then <HOME_POST> and <PL_POST> should be provided.
        epi = dfs["Episodes"]
        epi["original_index"] = epi.index

        # Remove any children with evidence of former UASC status
        header = pd.DataFrame()
        if "Header" in dfs:
            header_current = dfs["Header"]
            header = pd.concat((header, header_current), axis=0)
        elif "UASC" in dfs:
            uasc = dfs["UASC"]
            uasc = uasc.loc[
                uasc.drop("CHILD", axis="columns").notna().any(axis=1), ["CHILD"]
            ].copy()
            uasc.loc[:, "UASC"] = "1"
            header = pd.concat((header, uasc), axis=0)

        if "Header_last" in dfs:
            header = pd.concat((header, dfs["Header_last"]), axis=0)
        elif "UASC_last" in dfs:
            uasc = dfs["UASC_last"]
            uasc = uasc.loc[
                uasc.drop("CHILD", axis="columns").notna().any(axis=1), ["CHILD"]
            ].copy()
            uasc.loc[:, "UASC"] = "1"
            header = pd.concat((header, uasc), axis=0)

        if "UASC" in header.columns:
            header = header[header.UASC == "1"].drop_duplicates("CHILD")
            epi = epi.merge(
                header[["CHILD"]], how="left", on="CHILD", indicator=True
            ).query("_merge == 'left_only'")

        # Remove episodes where LS is V3/V4
        epi = epi.query("(~LS.isin(['V3','V4']))")

        # Remove episodes with postcodes filled in
        epi = epi.query("(HOME_POST.isna() | PL_POST.isna())")

        # Get error indices
        err_list = epi["original_index"].sort_values().tolist()

        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_episodes = pd.DataFrame(
        [
            {"CHILD": "111", "LS": "L1", "HOME_POST": "XX1", "PL_POST": "XX1"},  # 0
            {"CHILD": "222", "LS": "L1", "HOME_POST": "XX1", "PL_POST": pd.NA},  # 1
            {"CHILD": "222", "LS": "V3", "HOME_POST": pd.NA, "PL_POST": "XX1"},  # 2
            {"CHILD": "333", "LS": "L1", "HOME_POST": "XX1", "PL_POST": pd.NA},  # 3
            {"CHILD": "333", "LS": "V4", "HOME_POST": "XX1", "PL_POST": "XX1"},  # 4
            {"CHILD": "345", "LS": "L1", "HOME_POST": "XX1", "PL_POST": pd.NA},  # 5
            {"CHILD": "444", "LS": "L1", "HOME_POST": "XX1", "PL_POST": "XX1"},  # 6
            {"CHILD": "444", "LS": "V3", "HOME_POST": pd.NA, "PL_POST": pd.NA},  # 7
        ]
    )

    fake_header = pd.DataFrame(
        [
            {"CHILD": "111", "UASC": "1"},  # 0
            {"CHILD": "222", "UASC": "0"},  # 2
            {"CHILD": "333", "UASC": "0"},  # 4
            {"CHILD": "345", "UASC": "0"},  # 5
            {"CHILD": "444", "UASC": "0"},  # 6
        ]
    )
    fake_header_last = pd.DataFrame(
        [
            {"CHILD": "111", "UASC": "0"},  # 0
            {"CHILD": "222", "UASC": "0"},  # 2
            {"CHILD": "333", "UASC": "0"},  # 4
            {"CHILD": "345", "UASC": "0"},  # 5
            {"CHILD": "444", "UASC": "1"},  # 6
        ]
    )

    fake_dfs = {
        "Episodes": fake_episodes,
        "Header": fake_header,
        "Header_last": fake_header_last,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [1, 3, 5]}

    uasc_last = pd.DataFrame(
        [
            {"CHILD": "222", "DUC": "01/01/1990", "ETC": pd.NA},
            {"CHILD": "345", "DUC": pd.NA, "ETC": pd.NA},
        ]
    )

    fake_dfs = {"Episodes": fake_episodes, "UASC_last": uasc_last}

    result = validate(fake_dfs)

    assert result == {"Episodes": [3, 5]}



from typing import TypedDict


class UploadedFile(TypedDict):
    name: str
    file_content: bytes
    description: str


class UploadError(Exception):
    pass


class MissingMetadataError(KeyError):
    pass



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="635",
    message="There are entries for date of order and local authority code where previous permanence option was arranged but previous permanence code is Z1",
    affected_fields=["LA_PERM", "DATE_PERM", "PREV_PERM"],
    tables=["PrevPerm"],
)
def validate(dfs):
    if "PrevPerm" not in dfs:
        return {}
    else:
        prev_perm = dfs["PrevPerm"]
        # raise and error if either LA_PERM or DATE_PERM are present, yet PREV_PERM is absent.
        mask = (
            prev_perm["LA_PERM"].notna() | prev_perm["DATE_PERM"].notna()
        ) & prev_perm["PREV_PERM"].isna()

        error_locations = prev_perm.index[mask]
    return {"PrevPerm": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data_prevperm = pd.DataFrame(
        {
            "CHILD": ["2", "4", "5", "6", "7", "8"],
            "PREV_PERM": ["Z1", pd.NA, pd.NA, "Z1", pd.NA, "P1"],
            "LA_PERM": [pd.NA, "352", pd.NA, "352", pd.NA, "352"],
            "DATE_PERM": [pd.NA, pd.NA, "01/05/2000", pd.NA, pd.NA, "05/05/2020"],
        }
    )

    fake_dfs = {"PrevPerm": fake_data_prevperm}

    result = validate(fake_dfs)
    assert result == {"PrevPerm": [1, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="209",
    message="Child looked after is of school age and should not have an unknown Unique Pupil Number (UPN) code of UN1.",
    affected_fields=["UPN", "DOB"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        collection_start = dfs["metadata"]["collection_start"]
        # convert to datetime
        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )
        yr = collection_start.year - 1
        reference_date = pd.to_datetime(
            "31/08/" + str(yr), format="%d/%m/%Y", errors="coerce"
        )
        # If <DOB> >= 4 years prior to 31/08/YYYY then <UPN> should not be 'UN1' Note: YYYY in this instance refers to the year prior to the collection start (for collection year 2019-2020, it would be looking at the 31/08/2018).
        mask = (reference_date >= (header["DOB"] + pd.offsets.DateOffset(years=4))) & (
            header["UPN"] == "UN1"
        )
        # error locations
        error_locs_header = header.index[mask]
        return {"Header": error_locs_header.tolist()}


def test_validate():
    import pandas as pd

    fake_data_header = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "106", "107"],
            "UPN": [pd.NA, "H801200001001", "UN1", "UN2", pd.NA, "UN1", pd.NA],
            "DOB": [
                "01/01/2020",
                "11/11/2020",
                "03/10/2015",
                "11/11/2020",
                "01/01/2020",
                "11/11/2020",
                "01/02/2020",
            ],
        }
    )
    metadata = {
        "collection_start": "01/04/2025",
    }
    fake_dfs = {"Header": fake_data_header, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Header": [2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="511",
    message="If reporting that the number of person(s) adopting the looked after child is two adopters then the code should only be MM, FF or MF. MM = the adopting couple are both males; FF = the adopting couple are both females; MF = The adopting couple are male and female.",
    affected_fields=["NB_ADOPTR", "SEX_ADOPTR"],
    tables=["AD1"],
)
def validate(dfs):
    if "AD1" not in dfs:
        return {}

    else:
        AD1 = dfs["AD1"]

        mask = AD1["NB_ADOPTR"].astype(str).eq("2") & AD1["SEX_ADOPTR"].isin(
            ["M1", "F1"]
        )

        validation_error_mask = mask
        validation_error_locations = AD1.index[validation_error_mask]

        return {"AD1": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "NB_ADOPTR": ["2", "2", pd.NA, 2, "2", "1", 2, 1, 2],
            "SEX_ADOPTR": ["MM", "FF", "MM", "M1", "F1", "F1", "F1", "F1", pd.NA],
        }
    )

    fake_dfs = {"AD1": fake_data}

    result = validate(fake_dfs)

    assert result == {"AD1": [3, 4, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="388",
    message="Reason episode ceased is coded new episode begins, but there is no continuation episode.",
    affected_fields=["REC"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        df = dfs["Episodes"]
        df["DECOM"] = pd.to_datetime(df["DECOM"], format="%d/%m/%Y", errors="coerce")
        df["DEC"] = pd.to_datetime(df["DEC"], format="%d/%m/%Y", errors="coerce")

        df["DECOM"] = df["DECOM"].fillna(
            "01/01/1901"
        )  # Watch for potential future issues
        df = df.sort_values(["CHILD", "DECOM"])

        df["DECOM_NEXT_EPISODE"] = df.groupby(["CHILD"])["DECOM"].shift(-1)

        # The max DECOM for each child is also the one with no next episode
        # And we also add the skipna option
        # grouped_decom_by_child = df.groupby(['CHILD'])['DECOM'].idxmax(skipna=True)
        no_next = df.DECOM_NEXT_EPISODE.isna() & df.CHILD.notna()

        # Dataframe with the maximum DECOM removed
        max_decom_removed = df[~no_next]
        # Dataframe with the maximum DECOM only
        max_decom_only = df[no_next]

        # Case 1: If reason episode ceased is coded X1 there must be a subsequent episode
        #        starting on the same day.
        case1 = max_decom_removed[
            (max_decom_removed["REC"] == "X1")
            & (max_decom_removed["DEC"].notna())
            & (max_decom_removed["DECOM_NEXT_EPISODE"].notna())
            & (max_decom_removed["DEC"] != max_decom_removed["DECOM_NEXT_EPISODE"])
        ]

        # Case 2: If an episode ends but the child continues to be looked after, a new
        #        episode should start on the same day.The reason episode ceased code of
        #        the episode which ends must be X1.
        case2 = max_decom_removed[
            (max_decom_removed["REC"] != "X1")
            & (max_decom_removed["REC"].notna())
            & (max_decom_removed["DEC"].notna())
            & (max_decom_removed["DECOM_NEXT_EPISODE"].notna())
            & (max_decom_removed["DEC"] == max_decom_removed["DECOM_NEXT_EPISODE"])
        ]

        # Case 3: If a child ceases to be looked after reason episode ceased code X1 must
        #        not be used.
        case3 = max_decom_only[
            (max_decom_only["DEC"].notna()) & (max_decom_only["REC"] == "X1")
        ]

        mask_case1 = case1.index.tolist()
        mask_case2 = case2.index.tolist()
        mask_case3 = case3.index.tolist()

        mask = mask_case1 + mask_case2 + mask_case3

        mask.sort()
        return {"Episodes": mask}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DECOM": "01/06/2020",
                "DEC": "04/06/2020",
                "REC": "X1",
            },  # 0  Fails Case 1
            {
                "CHILD": "111",
                "DECOM": "05/06/2020",
                "DEC": "06/06/2020",
                "REC": "X1",
            },  # 1
            {
                "CHILD": "111",
                "DECOM": "06/06/2020",
                "DEC": "08/06/2020",
                "REC": "X1",
            },  # 2
            {
                "CHILD": "111",
                "DECOM": "08/06/2020",
                "DEC": "05/06/2020",
                "REC": "X1",
            },  # 3  Fails Case 3
            {
                "CHILD": "222",
                "DECOM": "05/06/2020",
                "DEC": "06/06/2020",
                "REC": pd.NA,
            },  # 4
            {
                "CHILD": "333",
                "DECOM": "06/06/2020",
                "DEC": "07/06/2020",
                "REC": "E11",
            },  # 5  Fails Case 2
            {"CHILD": "333", "DECOM": "07/06/2020", "DEC": pd.NA, "REC": pd.NA},  # 6
            {
                "CHILD": "444",
                "DECOM": "08/06/2020",
                "DEC": "09/06/2020",
                "REC": "X1",
            },  # 7
            {
                "CHILD": "444",
                "DECOM": "09/06/2020",
                "DEC": "10/06/2020",
                "REC": "E11",
            },  # 8
            {"CHILD": "444", "DECOM": "15/06/2020", "DEC": pd.NA, "REC": pd.NA},  # 9
            {
                "CHILD": "555",
                "DECOM": "11/06/2020",
                "DEC": "12/06/2020",
                "REC": "X1",
            },  # 10  Fails Case 3
            {
                "CHILD": "6666",
                "DECOM": "12/06/2020",
                "DEC": "13/06/2020",
                "REC": "X1",
            },  # 11
            {
                "CHILD": "6666",
                "DECOM": "13/06/2020",
                "DEC": "14/06/2020",
                "REC": "X1",
            },  # 12
            {
                "CHILD": "6666",
                "DECOM": "14/06/2020",
                "DEC": "15/06/2020",
                "REC": "X1",
            },  # 13
            {
                "CHILD": "6666",
                "DECOM": "15/06/2020",
                "DEC": "16/06/2020",
                "REC": "X1",
            },  # 14  Fails Case 3
            {
                "CHILD": "77777",
                "DECOM": "16/06/2020",
                "DEC": "17/06/2020",
                "REC": "X1",
            },  # 15
            {
                "CHILD": "77777",
                "DECOM": "17/06/2020",
                "DEC": "18/06/2020",
                "REC": "X1",
            },  # 16
            {"CHILD": "77777", "DECOM": "18/06/2020", "DEC": pd.NA, "REC": pd.NA},  # 17
            {
                "CHILD": "999",
                "DECOM": "31/06/2020",
                "DEC": pd.NA,
                "REC": pd.NA,
            },  # 18   Nonsense date, but should pass
            {
                "CHILD": "123",
                "DECOM": pd.NA,
                "DEC": pd.NA,
                "REC": pd.NA,
            },  # 19   Nonsense dates, but should pass
            {
                "CHILD": pd.NA,
                "DECOM": pd.NA,
                "DEC": pd.NA,
                "REC": pd.NA,
            },  # 20   Nonsense, but should pass
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 3, 5, 10, 14]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="611",
    message="Date of birth field is blank, but child is a mother.",
    affected_fields=["MOTHER", "MC_DOB"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        validation_error_mask = (
            header["MOTHER"].astype(str).isin(["1"]) & header["MC_DOB"].isna()
        )
        validation_error_locations = header.index[validation_error_mask]

        return {"Header": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "MOTHER": [1, "1", pd.NA, pd.NA, 1],
            "MC_DOB": ["01/01/2021", "19/02/2016", "dd/mm/yyyy", "31/31/19", pd.NA],
        }
    )

    fake_dfs = {"Header": fake_data}

    result = validate(fake_dfs)

    assert result == {"Header": [4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="NoE",
    message="This child has no episodes loaded for previous year even though child started to be looked after before this current year.",
    affected_fields=["DECOM"],
    tables=["Episodes", "Episodes_last"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Episodes_last" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        episodes_last = dfs["Episodes_last"]
        episodes_last["DECOM"] = pd.to_datetime(
            episodes_last["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        collection_start = pd.to_datetime(
            dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
        )

        episodes_before_year = episodes[episodes["DECOM"] < collection_start]

        episodes_merged = (
            episodes_before_year.reset_index()
            .merge(episodes_last, how="left", on=["CHILD"], indicator=True)
            .set_index("index")
        )

        episodes_not_matched = episodes_merged[episodes_merged["_merge"] == "left_only"]

        error_mask = episodes.index.isin(episodes_not_matched.index)

        error_locations = episodes.index[error_mask]

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103"],
            "DECOM": ["14/03/2021", "08/09/2021", "03/10/2020"],
        }
    )

    fake_data_prev = pd.DataFrame(
        {"CHILD": ["101", "102"], "DECOM": ["14/03/2021", "16/06/2019"]}
    )

    metadata = {"collection_start": "01/04/2021"}

    fake_dfs = {
        "Episodes": fake_data,
        "Episodes_last": fake_data_prev,
        "metadata": metadata,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [2]}



import pandas as pd

from lac_validator.datastore import merge_postcodes
from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="392d",
    message="Home and placement postcodes should not be same unless the placement type is P1.",
    affected_fields=["HOME_POST", "PL_POST", "PLACE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]

        not_p1 = episodes[(episodes["PLACE"] != "P1")]

        same_postcode = not_p1[not_p1["HOME_POST"] == not_p1["PL_POST"]]

        error_mask = episodes.index.isin(same_postcode.index)

        return {"Episodes": episodes.index[error_mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE": [
                "P2",
                "P1",
                "P2",
            ],
            "PL_POST": [
                "BN20",
                "BN20",
                "BN21",
            ],
            "HOME_POST": [
                "BN20",
                "BN20",
                "BN20",
            ],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {
        "Episodes": [
            0,
        ]
    }



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="571",
    message="The date that the child ceased to be missing or away from placement without authorisation is before the start or after the end of the collection year.",
    affected_fields=["MIS_END"],
    tables=["Missing"],
)
def validate(dfs):
    if "Missing" not in dfs:
        return {}
    else:
        missing = dfs["Missing"]
        collection_start = pd.to_datetime(
            dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
        )
        collection_end = pd.to_datetime(
            dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
        )

        missing["fMIS_END"] = pd.to_datetime(
            missing["MIS_END"], format="%d/%m/%Y", errors="coerce"
        )

        end_date_before_year = missing["fMIS_END"] < collection_start
        end_date_after_year = missing["fMIS_END"] > collection_end

        error_mask = end_date_before_year | end_date_after_year

        error_locations = missing.index[error_mask]

        return {"Missing": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "MIS_START": [
                "08/03/2020",
                "22/06/2020",
                pd.NA,
                "13/10/2021",
                "10/24/2021",
            ],
            "MIS_END": ["08/03/2020", pd.NA, "22/06/2020", "13/10/21", pd.NA],
        }
    )

    metadata = {"collection_start": "01/04/2020", "collection_end": "31/03/2020"}

    fake_dfs = {"Missing": fake_data, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Missing": [0, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="437",
    message="Reason episode ceased is child has died or is aged 18 or over but there are further episodes.",
    affected_fields=["REC"],
    tables=["Episodes"],
)

# !# potential false negatives, as this only operates on the current year's data
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]

        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        episodes.sort_values(["CHILD", "DECOM"], inplace=True)
        episodes[["NEXT_DECOM", "NEXT_CHILD"]] = episodes[["DECOM", "CHILD"]].shift(-1)

        # drop rows with missing DECOM as invalid/missing values can lead to errors
        episodes = episodes.dropna(subset=["DECOM"])

        ceased_e2_e15 = episodes["REC"].str.upper().astype(str).isin(["E2", "E15"])
        has_later_episode = episodes["CHILD"] == episodes["NEXT_CHILD"]

        error_mask = ceased_e2_e15 & has_later_episode

        error_locations = episodes.index[error_mask]

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        {
            "CHILD": [
                "101",
                "102",
                "102",
                "103",
                "103",
                "103",
                "104",
                "104",
                "104",
                "104",
                "105",
                "105",
            ],
            "DECOM": [
                "20/10/2021",
                "19/11/2021",
                "17/06/2021",
                "04/04/2020",
                "11/09/2020",
                "07/05/2021",
                "15/02/2020",
                "09/04/2020",
                "24/09/2020",
                "30/06/2021",
                pd.NA,
                "20/12/2020",
            ],
            "REC": [
                "E2",
                "E15",
                "X1",
                "X1",
                "E15",
                pd.NA,
                "X1",
                "E4a",
                "X1",
                pd.NA,
                "E2",
                "X1",
            ],
        }
    )

    fake_dfs = {"Episodes": fake_data_episodes}

    result = validate(fake_dfs)

    assert result == {"Episodes": [4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="445",
    message="D1 is not a valid code for episodes starting after December 2005.",
    affected_fields=["LS", "DECOM"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        max_decom_allowed = pd.to_datetime(
            "31/12/2005", format="%d/%m/%Y", errors="coerce"
        )

        mask = episodes["LS"].eq("D1") & (episodes["DECOM"] > max_decom_allowed)
        validation_error_mask = mask
        validation_error_locations = episodes.index[validation_error_mask]

        return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "LS": ["D1", "D1", "D1", "D1", "D1", "C1"],
            "DECOM": [
                "01/11/2005",
                "31/12/2005",
                pd.NA,
                "20/01/2006",
                "01/10/2012",
                "20/02/2005",
            ],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="382",
    message="A child receiving respite care cannot be in a temporary placement.",
    affected_fields=["LS", "PLACE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        err_list = epi.query(
            "LS.isin(['V3', 'V4']) & PLACE.isin(['T0', 'T1', 'T2', 'T3', 'T4'])"
        ).index.tolist()
        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"PLACE": "T0", "LS": "V3"},  # 0
            {"PLACE": "R4", "LS": "V3"},  # 1
            {"PLACE": "T1", "LS": "X3"},  # 2
            {"PLACE": "T2", "LS": "V4"},  # 3
            {"PLACE": "T3", "LS": "V4"},  # 4
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="352",
    message="Child who started to be looked after was aged 18 or over.",
    affected_fields=["DECOM", "RNE"],
    tables=["Header", "Episodes"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    if "Episodes" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        episodes = dfs["Episodes"]

        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        header["DOB18"] = header["DOB"] + pd.DateOffset(years=18)

        episodes_merged = (
            episodes.reset_index()
            .merge(
                header,
                how="left",
                on=["CHILD"],
                suffixes=("", "_header"),
                indicator=True,
            )
            .set_index("index")
        )

        care_start = episodes_merged["RNE"].str.upper().astype(str).isin(["S"])
        started_over_18 = episodes_merged["DOB18"] <= episodes_merged["DECOM"]

        error_mask = care_start & started_over_18

        error_locations = episodes_merged.index[error_mask].unique()

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "101", "102", "103"],
            "RNE": ["S", "S", "X1", pd.NA, "S"],
            "DECOM": ["16/03/2021", "17/06/2020", "20/03/2020", pd.NA, "23/08/2020"],
        }
    )

    fake_data_child = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "103"],
            "DOB": ["16/03/2005", "23/09/2002", "31/12/2000", "31/12/2000"],
        }
    )

    fake_dfs = {"Episodes": fake_data, "Header": fake_data_child}

    result = validate(fake_dfs)

    assert result == {"Episodes": [4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="434",
    message="Reason for new episode is that child's legal status has changed but not the placement, but this is not reflected in the episode data.",
    affected_fields=["RNE", "LS", "PLACE", "PL_POST", "URN", "PLACE_PROVIDER"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        # create columns of previous values

        cols = ["PLACE", "PL_POST", "URN", "PLACE_PROVIDER"]

        episodes = episodes.sort_values(["CHILD", "DECOM"])

        error_mask = (
            (episodes["CHILD"] == episodes["CHILD"].shift(1))
            & (episodes["RNE"] == "L")
            & (
                (episodes["LS"] == episodes["LS"].shift(1))
                | (episodes[cols].fillna("") != episodes[cols].shift(1).fillna("")).any(
                    axis=1
                )
            )
        )
        # error locations
        error_locs = episodes.index[error_mask].to_list()
        return {"Episodes": error_locs}


def test_validate():
    import pandas as pd

    fake_eps = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DECOM": "01/01/2020",
                "RNE": "L",
                "LS": "--",
                "PL_POST": "ooo",
                "URN": "---",
            },  # 0
            {
                "CHILD": "111",
                "DECOM": "01/02/2020",
                "RNE": "P",
                "LS": "--",
                "PL_POST": "ooo",
                "URN": "---",
            },  # 1
            {
                "CHILD": "111",
                "DECOM": "01/12/2020",
                "RNE": "L",
                "LS": "--",
                "PL_POST": "ooo",
                "URN": "---",
            },
            # 2 fail: LS same as [3]
            {
                "CHILD": "111",
                "DECOM": "01/03/2020",
                "RNE": "P",
                "LS": "--",
                "PL_POST": "ooo",
                "URN": "---",
            },  # 3
            {
                "CHILD": "222",
                "DECOM": "01/01/2020",
                "RNE": "L",
                "LS": "--",
                "PL_POST": "---",
                "URN": "---",
            },  # 4
            {
                "CHILD": "222",
                "DECOM": "01/02/2020",
                "RNE": "L",
                "LS": "xx",
                "PL_POST": "---",
                "URN": "---",
            },  # 5
            {
                "CHILD": "222",
                "DECOM": "01/03/2020",
                "RNE": "L",
                "LS": "oo",
                "PL_POST": "---",
                "URN": "xxx",
            },  # 6 fail: URN
            {
                "CHILD": "222",
                "DECOM": "01/04/2020",
                "RNE": "L",
                "LS": "oo",
                "PL_POST": "xxx",
                "URN": "xxx",
            },
            # 7 fail: PL_POST
            {
                "CHILD": "333",
                "DECOM": "01/03/2020",
                "RNE": "P",
                "LS": "C1",
                "PL_POST": "---",
                "URN": pd.NA,
            },  # 8
            {
                "CHILD": "333",
                "DECOM": "01/04/2020",
                "RNE": "L",
                "LS": "C2",
                "PL_POST": "---",
                "URN": pd.NA,
            },  # 9
        ]
    )
    fake_eps["PLACE"] = "---"
    fake_eps["PLACE_PROVIDER"] = "---"
    fake_dfs = {"Episodes": fake_eps}

    result = validate(fake_dfs)
    assert result == {"Episodes": [2, 6, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="351",
    message="Child was over 25 at the start of the current collection year.",
    affected_fields=[
        "DOB",
    ],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}

    else:
        header = dfs["Header"]
        collection_start = dfs["metadata"]["collection_start"]

        # Convert from string to date to appropriate format
        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )

        mask = collection_start > (header["DOB"] + pd.DateOffset(years=25))
        # error locations
        header_error_locs = header.index[mask]

        return {"Header": header_error_locs.tolist()}


def test_validate():
    import pandas as pd

    fake_data_header = pd.DataFrame(
        {
            "CHILD": [
                "101",
                "102",
                "103",
                "104",
                "105",
            ],
            "DOB": [
                pd.NA,
                "11/11/2020",
                "03/10/1995",
                "01/04/2000",
                "01/01/1999",
            ],
        }
    )
    metadata = {
        "collection_start": "01/04/2021",
    }
    fake_dfs = {"Header": fake_data_header, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Header": [2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="560",
    message="Date of decision that the child should be placed for adoption this year is different from that recorded last year but the decision to place the child for adoption did not change.",
    affected_fields=["DATE_PLACED", "DATE_PLACED_CEASED"],
    tables=["PlacedAdoption", "PlacedAdoption_last"],
)
def validate(dfs):
    if "PlacedAdoption" not in dfs or "PlacedAdoption_last" not in dfs:
        return {}
    else:
        placed_adoption = dfs["PlacedAdoption"]
        pa_last = dfs["PlacedAdoption_last"]

        # prepare to merge
        placed_adoption.reset_index(inplace=True)
        pa_last.reset_index(inplace=True)
        merged = placed_adoption.merge(
            pa_last, how="inner", on="CHILD", suffixes=["_now", "_last"]
        )

        # If <CURRENT_COLLECTION_YEAR> -1 <DATE_PLACED> has been provided and <DATE_PLACED_CEASED> is Null then <CURRENT_COLLECTION_YEAR> <DATE_PLACED> should = <CURRENT_COLLECTION_YEAR> -1 <DATE_PLACED>
        mask = (
            merged["DATE_PLACED_last"].notna()
            & merged["DATE_PLACED_CEASED_last"].isna()
            & (merged["DATE_PLACED_now"] != merged["DATE_PLACED_last"])
        )
        # error locations
        error_locs = merged.loc[mask, "index_now"]
        return {"PlacedAdoption": error_locs.tolist()}


def test_validate():
    import pandas as pd

    fake_placed_adoption = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DATE_PLACED_CEASED": "26/05/2000",
                "DATE_PLACED": "26/05/2019",
            },  # 0 --- FAIL
            {
                "CHILD": 102,
                "DATE_PLACED_CEASED": "01/07/2018",
                "DATE_PLACED": "26/05/2000",
            },  # 1
            {
                "CHILD": 103,
                "DATE_PLACED_CEASED": "26/05/2000",
                "DATE_PLACED": pd.NA,
            },  # 2
            {
                "CHILD": 104,
                "DATE_PLACED_CEASED": "26/05/2017",
                "DATE_PLACED": "01/02/2016",
            },  # 3
            {
                "CHILD": 105,
                "DATE_PLACED_CEASED": pd.NA,
                "DATE_PLACED": "27/05/2019",
            },  # 4 --- FAIL
            {
                "CHILD": 106,
                "DATE_PLACED_CEASED": "34rd/Dec/-1000A.D",
                "DATE_PLACED": "34rd/Dec/-1000A.D",
            },  # 5
            {
                "CHILD": 107,
                "DATE_PLACED_CEASED": "different",
                "DATE_PLACED": "also different",
            },  # 6
        ]
    )
    fake_pa_last = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DATE_PLACED_CEASED": pd.NA,
                "DATE_PLACED": "26/05/2000",
            },  # 0
            {
                "CHILD": 102,
                "DATE_PLACED_CEASED": "01/07/2018",
                "DATE_PLACED": "26/05/2000",
            },  # 1
            {
                "CHILD": 103,
                "DATE_PLACED_CEASED": "26/05/2000",
                "DATE_PLACED": pd.NA,
            },  # 2
            {
                "CHILD": 104,
                "DATE_PLACED_CEASED": "26/05/2017",
                "DATE_PLACED": "01/02/2016",
            },  # 3
            {
                "CHILD": 105,
                "DATE_PLACED_CEASED": pd.NA,
                "DATE_PLACED": "26/05/2019",
            },  # 4
            {
                "CHILD": 106,
                "DATE_PLACED_CEASED": pd.NA,
                "DATE_PLACED": "34rd/Dec/-1000A.D",
            },  # 5
            {
                "CHILD": 107,
                "DATE_PLACED_CEASED": "34rd/Dec/-1000A.D",
                "DATE_PLACED": "20/20/2020",
            },  # 6
        ]
    )
    fake_dfs = {
        "PlacedAdoption": fake_placed_adoption,
        "PlacedAdoption_last": fake_pa_last,
    }

    result = validate(fake_dfs)
    assert result == {"PlacedAdoption": [0, 4]}



from pathlib import Path

from lac_validator.rule_engine import RuleDefinition, YearConfig
from lac_validator.rules.lac2023_24 import registry as prev_registry
from lac_validator.rules.ruleset_utils import (
    extract_validator_functions,
    update_validator_functions,
)

files = Path(__file__).parent.glob("*.py")
this_year_validator_funcs: dict[str, RuleDefinition] = extract_validator_functions(
    files
)
# if any rules need to be deleted, add their codes as strings into del_list
del_list: list[str] = [
    "217t",
    "230",
    "370",
    "445",
    "451",
    "555",
    "556",
    "SW01STG1",
    "SW03STG1",
    "SW11aSTG2",
    "SW14STG2",
    "SW15STG2",
]
this_year_config = YearConfig(
    deleted=del_list, added_or_modified=this_year_validator_funcs
)

registry = update_validator_functions(prev_registry, this_year_config)
__all__ = ["registry"]



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="192",
    message="Child has been identified as having a substance misuse problem but the additional item on whether an intervention was received has been left blank.",
    affected_fields=["SUBSTANCE_MISUSE", "INTERVENTION_RECEIVED"],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]

        misuse = oc2["SUBSTANCE_MISUSE"].astype(str) == "1"
        intervention_blank = oc2["INTERVENTION_RECEIVED"].isna()

        error_mask = misuse & intervention_blank
        validation_error_locations = oc2.index[error_mask]

        return {"OC2": validation_error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SUBSTANCE_MISUSE": [0, pd.NA, 0, "1", 1, "1"],
            "INTERVENTION_RECEIVED": [0, pd.NA, "1", 1, pd.NA, pd.NA],
        }
    )

    fake_dfs = {"OC2": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC2": [4, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="433",
    message="The reason for new episode suggests that this is a continuation episode, but the episode does not start on the same day as the last episode finished.",
    affected_fields=["RNE", "DECOM"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        episodes["DECOM_dt"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC_dt"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )

        episodes["original_index"] = episodes.index
        episodes.sort_values(["CHILD", "DECOM_dt", "DEC_dt"], inplace=True)
        episodes[["PREVIOUS_DEC", "PREVIOUS_CHILD"]] = episodes[["DEC", "CHILD"]].shift(
            1
        )

        rne_is_ongoing = (
            episodes["RNE"].str.upper().astype(str).isin(["P", "L", "T", "U", "B"])
        )
        date_mismatch = episodes["PREVIOUS_DEC"] != episodes["DECOM"]
        missing_date = episodes["PREVIOUS_DEC"].isna() | episodes["DECOM"].isna()
        same_child = episodes["PREVIOUS_CHILD"] == episodes["CHILD"]

        error_mask = rne_is_ongoing & (date_mismatch | missing_date) & same_child

        error_locations = episodes["original_index"].loc[error_mask].sort_values()

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        [
            {
                "CHILD": "101",
                "DECOM": "20/10/2021",
                "RNE": "S",
                "DEC": "20/11/2021",
            },  # 0: Ignore
            {
                "CHILD": "102",
                "DECOM": "19/11/2021",
                "RNE": "P",
                "DEC": pd.NA,
            },  # 1 [102:2nd]
            {
                "CHILD": "102",
                "DECOM": "17/06/2021",
                "RNE": "P",
                "DEC": "19/11/2021",
            },  # 2 [102:1st]
            {
                "CHILD": "103",
                "DECOM": "04/04/2020",
                "RNE": "B",
                "DEC": "12/09/2020",
            },  # 3 [103:1st]
            {
                "CHILD": "103",
                "DECOM": "11/09/2020",
                "RNE": "B",
                "DEC": "06/05/2021",
            },  # 4 [103:2nd] ]Fail!
            {
                "CHILD": "103",
                "DECOM": "07/05/2021",
                "RNE": "B",
                "DEC": pd.NA,
            },  # 5 [103:3rd] Fail!
        ]
    )

    fake_dfs = {"Episodes": fake_data_episodes}

    result = validate(fake_dfs)

    assert result == {"Episodes": [4, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="384",
    message="A child receiving respite care cannot be in a long-term foster placement ",
    affected_fields=["PLACE", "LS"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        # Where <LS> = 'V3' or 'V4' then <PL> must not be 'U1' or 'U4'
        mask = ((episodes["LS"] == "V3") | (episodes["LS"] == "V4")) & (
            (episodes["PLACE"] == "U1") | (episodes["PLACE"] == "U4")
        )
        error_locations = episodes.index[mask]
        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "11",
                "PLACE": "U1",
                "LS": "V4",
            },  # 0
            {
                "CHILD": "202",
                "PLACE": "x",
                "LS": "D1",
            },  # 1
            {
                "CHILD": "3003",
                "PLACE": "U4",
                "LS": "V3",
            },  # 2
            {
                "CHILD": "40004",
                "PLACE": "P1",
                "LS": "V3",
            },  # 3
            {
                "CHILD": "5005",
                "PLACE": "A5",
                "LS": "x",
            },  # 4
            {
                "CHILD": "606",
                "PLACE": "A6",
                "LS": "V4",
            },  # 5
            {"CHILD": "77", "PLACE": "x", "LS": "x"},  # 6
        ]
    )
    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)
    assert result == {"Episodes": [0, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="625",
    message="Date of birth of the first child is beyond the end of this reporting year or the date the child ceased to be looked after.",
    affected_fields=["MC_DOB", "DEC"],
    tables=["Episodes", "Header"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Header" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        header = dfs["Header"]
        collection_end = dfs["metadata"]["collection_end"]

        # datetime conversion
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )
        header["MC_DOB"] = pd.to_datetime(
            header["MC_DOB"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        # prepare to merge
        header.reset_index(inplace=True)
        episodes.reset_index(inplace=True)

        # if nans aren't dropped, idxmax() won't work. we can do this since dropping nan DECs doesn't affect the rule logic.
        dec_only_eps = episodes[["CHILD", "DEC"]].dropna()
        # latest episodes
        eps_last_indices = dec_only_eps.groupby("CHILD")["DEC"].idxmax()
        latest_episodes = episodes[episodes.index.isin(eps_last_indices)]
        merged = latest_episodes.merge(
            header, on="CHILD", how="left", suffixes=["_eps", "_er"]
        )
        # If provided <MC_DOB> must not be > <COLLECTION_END> or <DEC> of latest episode
        mask = (merged["MC_DOB"] > collection_end) | (merged["MC_DOB"] > merged["DEC"])
        header_error_locs = merged.loc[mask, "index_er"]
        eps_error_locs = merged.loc[mask, "index_eps"]
        return {
            "Header": header_error_locs.unique().tolist(),
            "Episodes": eps_error_locs.tolist(),
        }


def test_validate():
    import pandas as pd

    fake_data_header = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103"],
            "MC_DOB": ["01/11/2021", "19/12/2016", pd.NA],
            # 0 MC_DOB > collection_end FAIL
            # 1 MC_DOB < collection_end but > DEC of latest episode FAIL
            # 2 MC_DOB not provided IGNORE
        }
    )
    fake_data_episodes = pd.DataFrame(
        {
            "CHILD": ["101", "101", "102", "102", "103", "103", "103"],
            "DEC": [
                "02/12/2021",
                "11/11/2012",
                "03/10/2014",
                "11/11/2015",
                "01/01/2020",
                "11/11/2020",
                "01/02/2020",
            ],
        }
    )
    metadata = {"collection_end": "31/03/2021"}
    fake_dfs = {
        "metadata": metadata,
        "Episodes": fake_data_episodes,
        "Header": fake_data_header,
    }

    result = validate(fake_dfs)
    assert result == {"Episodes": [0, 3], "Header": [0, 1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="158",
    message="If a child has been recorded as receiving an intervention for their substance misuse problem, then the additional item on whether an intervention was offered should be left blank.",
    affected_fields=["INTERVENTION_RECEIVED", "INTERVENTION_OFFERED"],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}

    else:
        oc2 = dfs["OC2"]

        error_mask = (
            oc2["INTERVENTION_RECEIVED"].astype(str).eq("1")
            & oc2["INTERVENTION_OFFERED"].notna()
        )

        error_locations = oc2.index[error_mask]

        return {"OC2": error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "INTERVENTION_RECEIVED": ["1", pd.NA, "0", "1", "1", "0"],
            "INTERVENTION_OFFERED": [pd.NA, "1", "0", "1", "0", "1"],
        }
    )

    fake_dfs = {"OC2": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC2": [3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="184",
    message="Date of decision that a child should be placed for adoption is before the child was born.",
    affected_fields=["DATE_PLACED", "DOB"],  # PlacedAdoptino  # Header
    tables=["Header", "PlacedAdoption"],
)
def validate(dfs):
    if "Header" not in dfs or "PlacedAdoption" not in dfs:
        return {}
    else:
        child_record = dfs["Header"]
        placed_for_adoption = dfs["PlacedAdoption"]

        all_data = placed_for_adoption.reset_index().merge(
            child_record, how="left", on="CHILD", suffixes=[None, "_P4A"]
        )

        all_data["DATE_PLACED"] = pd.to_datetime(
            all_data["DATE_PLACED"], format="%d/%m/%Y", errors="coerce"
        )
        all_data["DOB"] = pd.to_datetime(
            all_data["DOB"], format="%d/%m/%Y", errors="coerce"
        )

        mask = (all_data["DATE_PLACED"] >= all_data["DOB"]) | all_data[
            "DATE_PLACED"
        ].isna()

        validation_error = ~mask

        validation_error_locations = all_data[validation_error]["index"].unique()

        return {"PlacedAdoption": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data_header = pd.DataFrame(
        {
            "CHILD": ["111", "112", "113", "114"],
            "DOB": ["01/10/2017", "31/05/2018", "10/03/2019", "19/08/2020"],
        }
    )
    fake_data_placed = pd.DataFrame(
        {
            "CHILD": ["111", "112", "113", "114"],
            "DATE_PLACED": ["01/10/2017", "10/02/2019", "10/02/2019", pd.NA],
        }
    )

    fake_dfs = {"Header": fake_data_header, "PlacedAdoption": fake_data_placed}

    result = validate(fake_dfs)

    assert result == {"PlacedAdoption": [2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="567",
    message="The date that the missing episode or episode that the child was away from placement without authorisation ended is before the date that it started.",
    affected_fields=["MIS_START", "MIS_END"],
    tables=["Missing"],
)
def validate(dfs):
    if "Missing" not in dfs:
        return {}
    else:
        mis = dfs["Missing"]
        mis["MIS_START"] = pd.to_datetime(
            mis["MIS_START"], format="%d/%m/%Y", errors="coerce"
        )
        mis["MIS_END"] = pd.to_datetime(
            mis["MIS_END"], format="%d/%m/%Y", errors="coerce"
        )

        mis_error = mis[mis["MIS_START"] > mis["MIS_END"]]

        return {"Missing": mis_error.index.to_list()}


def test_validate():
    import pandas as pd

    fake_mis = pd.DataFrame(
        [
            {"MIS_START": "01/06/2020", "MIS_END": "05/06/2020"},  # 0
            {"MIS_START": "02/06/2020", "MIS_END": pd.NA},  # 1
            {"MIS_START": "03/06/2020", "MIS_END": "01/06/2020"},  # 2 Fails
            {"MIS_START": "04/06/2020", "MIS_END": "02/06/2020"},  # 3 Fails
            {"MIS_START": pd.NA, "MIS_END": "05/06/2020"},  # 4
        ]
    )

    fake_dfs = {"Missing": fake_mis}

    result = validate(fake_dfs)

    assert result == {"Missing": [2, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="566",
    message="The date that the child"
    + chr(39)
    + "s episode of being missing or away from placement without authorisation ended has been completed but whether the child was missing or away without authorisation has not been completed.",
    affected_fields=["MISSING", "MIS_END"],
    tables=["Missing"],
)
def validate(dfs):
    if "Missing" not in dfs:
        return {}
    else:
        mis = dfs["Missing"]
        error_mask = mis["MISSING"].isna() & mis["MIS_END"].notna()
        return {"Missing": mis.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"MISSING": "M", "MIS_END": pd.NA},  # 0
            {"MISSING": pd.NA, "MIS_END": "07/02/2020"},  # 1  Fails
            {"MISSING": "A", "MIS_END": "03/02/2020"},  # 2
            {"MISSING": pd.NA, "MIS_END": pd.NA},  # 3
            {"MISSING": "M", "MIS_END": "01/02/2020"},  # 4
            {"MISSING": pd.NA, "MIS_END": "13/02/2020"},  # 5  Fails
        ]
    )

    fake_dfs = {"Missing": fake_data}

    assert validate(fake_dfs) == {"Missing": [1, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="334",
    message="Date child started to be looked after in latest period of care must be on or prior to the date should be placed for adoption. ",
    affected_fields=["DATE_INT", "DECOM", "RNE"],
    tables=["Episodes", "AD1"],
)
def validate(dfs):
    if "Episodes" not in dfs or "AD1" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        ad1 = dfs["AD1"]

        # to datetime
        ad1["DATE_INT"] = pd.to_datetime(
            ad1["DATE_INT"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        # select the earliest episodes with RNE =  S
        eps_rne = episodes[episodes["RNE"] == "S"]
        last_eps_idxs = eps_rne.groupby("CHILD")["DECOM"].idxmax()
        last_eps = eps_rne.loc[last_eps_idxs]

        # prepare to merge
        ad1.reset_index(inplace=True)
        last_eps.reset_index(inplace=True)
        merged = last_eps.merge(ad1, how="left", on="CHILD", suffixes=["_eps", "_ad1"])

        # <DATE_PLACED> cannot be prior to <DECOM> of the first episode with <RNE> = 'S'
        mask = merged["DATE_INT"] < merged["DECOM"]

        eps_error_locs = merged.loc[mask, "index_eps"]
        ad1_error_locs = merged.loc[mask, "index_ad1"]

        return {"Episodes": eps_error_locs.tolist(), "AD1": ad1_error_locs.tolist()}


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        [
            {"CHILD": 101, "DECOM": "01/03/1980", "RNE": "S"},  # 0 fail
            {"CHILD": 102, "DECOM": "01/03/1980", "RNE": "o"},
            {
                "CHILD": 102,
                "DECOM": "01/03/1980",
                "RNE": "S",
            },  # 2 ignore DATE_PLACED is nan
            {"CHILD": 103, "DECOM": "01/02/1970", "RNE": "o"},
            {"CHILD": 103, "DECOM": "01/03/1979", "RNE": "S"},
            {"CHILD": 103, "DECOM": "01/01/1980", "RNE": "S"},  # 5 pass
            {"CHILD": 104, "DECOM": "01/03/1979", "RNE": "o"},  # ignore no RNE is S
            {"CHILD": 104, "DECOM": "01/01/1981", "RNE": "o"},
            {"CHILD": 105, "DECOM": "01/03/1979", "RNE": "o"},
            {"CHILD": 105, "DECOM": "01/01/2020", "RNE": "S"},  # 9 fail
            {"CHILD": 105, "DECOM": "26/05/2021", "RNE": "o"},
        ]
    )
    fake_placed_ad1 = pd.DataFrame(
        [
            {"CHILD": 101, "DATE_INT": "26/05/1978"},  # 0 fail
            {"CHILD": 102, "DATE_INT": pd.NA},  # 1
            {"CHILD": 103, "DATE_INT": "26/05/1981"},  # 2 pass
            {"CHILD": 104, "DATE_INT": "01/02/1960"},  # 3
            {"CHILD": 105, "DATE_INT": "26/05/2019"},  # 4 fail
        ]
    )
    fake_dfs = {"Episodes": fake_data_episodes, "AD1": fake_placed_ad1}

    result = validate(fake_dfs)
    assert result == {"Episodes": [0, 9], "AD1": [0, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="153",
    message="All data items relating to a child's activity or accommodation after leaving care must be coded or left blank.",
    affected_fields=["IN_TOUCH", "ACTIV", "ACCOM"],
    tables=["OC3"],
)
def validate(dfs):
    if "OC3" not in dfs:
        return {}

    oc3 = dfs["OC3"]

    oc3_not_na = oc3["IN_TOUCH"].notna() & oc3["ACTIV"].notna() & oc3["ACCOM"].notna()

    oc3_all_na = oc3["IN_TOUCH"].isna() & oc3["ACTIV"].isna() & oc3["ACCOM"].isna()

    validation_error = ~oc3_not_na & ~oc3_all_na

    validation_error_locations = oc3.index[validation_error]

    return {"OC3": validation_error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "IN_TOUCH": ["XXX", pd.NA, "XXX", pd.NA, pd.NA, "XXX"],
            "ACTIV": ["XXX", pd.NA, pd.NA, "XXX", pd.NA, "XXX"],
            "ACCOM": ["XXX", pd.NA, pd.NA, pd.NA, "XXX", pd.NA],
        }
    )

    fake_dfs = {"OC3": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC3": [2, 3, 4, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="431",
    message="The reason for new episode is started to be looked after, but the previous episode ended on the same day.",
    affected_fields=["RNE", "DECOM"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi["DEC"] = pd.to_datetime(epi["DEC"], format="%d/%m/%Y", errors="coerce")
        epi.sort_values(["CHILD", "DECOM"], inplace=True)

        epi.reset_index(inplace=True)
        epi.reset_index(inplace=True)
        epi["LAG_INDEX"] = epi["level_0"].shift(-1)

        m_epi = epi.merge(
            epi,
            how="inner",
            left_on="level_0",
            right_on="LAG_INDEX",
            suffixes=["", "_PREV"],
        )

        m_epi = m_epi[(m_epi["CHILD"] == m_epi["CHILD_PREV"]) & (m_epi["RNE"] == "S")]
        error_mask = m_epi["DECOM"] <= m_epi["DEC_PREV"]
        error_list = m_epi["index"][error_mask].to_list()
        error_list.sort()
        return {"Episodes": error_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "RNE": "S",
                "DECOM": "01/06/2020",
                "DEC": "04/06/2020",
            },  # 0
            {
                "CHILD": "111",
                "RNE": "S",
                "DECOM": "05/06/2020",
                "DEC": "06/06/2020",
            },  # 1
            {
                "CHILD": "111",
                "RNE": "T",
                "DECOM": "06/06/2020",
                "DEC": "08/06/2020",
            },  # 2
            {
                "CHILD": "111",
                "RNE": "S",
                "DECOM": "08/06/2020",
                "DEC": "05/06/2020",
            },  # 3  Fails
            {
                "CHILD": "222",
                "RNE": "S",
                "DECOM": "05/06/2020",
                "DEC": "06/06/2020",
            },  # 4
            {
                "CHILD": "333",
                "RNE": "S",
                "DECOM": "06/06/2020",
                "DEC": "07/06/2020",
            },  # 5
            {"CHILD": "333", "RNE": "S", "DECOM": "10/06/2020", "DEC": pd.NA},  # 6
            {
                "CHILD": "444",
                "RNE": "S",
                "DECOM": "08/06/2020",
                "DEC": "09/06/2020",
            },  # 7
            {
                "CHILD": "444",
                "RNE": "S",
                "DECOM": "09/06/2020",
                "DEC": "10/06/2020",
            },  # 8  Fails
            {"CHILD": "444", "RNE": "S", "DECOM": "15/06/2020", "DEC": pd.NA},  # 9
            {
                "CHILD": "555",
                "RNE": "S",
                "DECOM": "11/06/2020",
                "DEC": "12/06/2020",
            },  # 10
            {
                "CHILD": "6666",
                "RNE": "S",
                "DECOM": "12/06/2020",
                "DEC": "13/06/2020",
            },  # 11
            {
                "CHILD": "6666",
                "RNE": "S",
                "DECOM": "13/06/2020",
                "DEC": "14/06/2020",
            },  # 12 Fails
            {
                "CHILD": "6666",
                "RNE": "S",
                "DECOM": "14/06/2020",
                "DEC": "15/06/2020",
            },  # 13 Fails
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [3, 8, 12, 13]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1000",
    message="This child is recorded as having died in care and therefore should not have the care leaver information completed. [NOTE: This only tests the current and previous year data loaded into the tool]",
    affected_fields=["IN_TOUCH", "ACTIV", "ACCOM"],
    tables=["Episodes", "OC3"],
)
def validate(dfs):
    if "Episodes" not in dfs or "OC3" not in dfs:
        return {}

    else:
        episodes = dfs["Episodes"]
        oc3 = dfs["OC3"]

        episodes_ended_e2 = episodes["REC"].str.upper().astype(str).isin(["E2"])
        episodes = episodes.loc[episodes_ended_e2]

        if "Episodes_last" in dfs:
            episodes_last = dfs["Episodes_last"]
            episodes_last_ended_e2 = (
                episodes_last["REC"].str.upper().astype(str).isin(["E2"])
            )
            episodes_last = episodes_last.loc[episodes_last_ended_e2]
            has_previous_e2 = oc3["CHILD"].isin(episodes_last["CHILD"])
            has_current_e2 = oc3["CHILD"].isin(episodes["CHILD"])
            error_mask = has_current_e2 | has_previous_e2
        else:
            has_current_e2 = oc3["CHILD"].isin(episodes["CHILD"])
            error_mask = has_current_e2

        validation_error_locations = oc3.index[error_mask]

        return {"OC3": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_episodes_prev = pd.DataFrame(
        {
            "CHILD": ["101", "101", "102", "102", "103"],
            "REC": ["X1", pd.NA, pd.NA, "E2", "E4a"],
        }
    )

    fake_episodes = pd.DataFrame({"CHILD": ["104", "105"], "REC": ["X1", "E2"]})

    fake_oc3 = pd.DataFrame({"CHILD": ["101", "102", "103", "104", "105"]})

    fake_dfs = {
        "Episodes": fake_episodes,
        "Episodes_last": fake_episodes_prev,
        "OC3": fake_oc3,
    }
    result = validate(fake_dfs)
    assert result == {"OC3": [1, 4]}

    fake_dfs = {"Episodes": fake_episodes, "OC3": fake_oc3}
    result = validate(fake_dfs)
    assert result == {"OC3": [4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="217",
    message="Children who are placed for adoption with current foster carers (placement types A3 or A5) must have a reason for new episode of S, T or U.",
    affected_fields=["PLACE", "DECOM", "RNE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        max_decom_allowed = pd.to_datetime(
            "01/04/2015", format="%d/%m/%Y", errors="coerce"
        )
        reason_new_ep = ["S", "T", "U"]
        place_codes = ["A3", "A5"]

        mask = (
            episodes["PLACE"].isin(place_codes)
            & (episodes["DECOM"] >= max_decom_allowed)
        ) & ~episodes["RNE"].isin(reason_new_ep)

        validation_error_mask = mask
        validation_error_locations = episodes.index[validation_error_mask]

        return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE": ["A3", "A5", "A3", pd.NA, "U1", "A3", "A5", "A5"],
            "DECOM": [
                "01/04/2015",
                "31/12/2015",
                "20/01/2016",
                pd.NA,
                "01/10/2017",
                "20/02/2016",
                "01/01/2017",
                "01/04/2013",
            ],
            "RNE": ["S", "T", "U", pd.NA, "X", "X", "X", "X"],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [5, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW16cSTG2",
    message="The social worker reason episode changed does not match the open episode at the end of last year.",
    affected_fields=["SW_REASON"],
    tables=["SWEpisodes", "SWEpisodes_last"],
)
def validate(dfs):
    if ("SWEpisodes" not in dfs) | ("SWEpisodes_last" not in dfs):
        return {}
    else:
        SWE = dfs["SWEpisodes"]
        SWE_prev = dfs["SWEpisodes_last"]

        # If present, where there is a previous collection year’s final social worker episode, and <SW_DEC> not provided,
        # then first social worker episode of this year’s collection <SW_REASON> be the same as the previous episode

        SWE_prev["SW_DECOM"] = pd.to_datetime(
            SWE_prev["SW_DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        prev_ordered = SWE_prev.sort_values(by=["SW_DECOM"], ascending=False)
        prev_ordered = prev_ordered.drop_duplicates(["CHILD"], keep="first")

        SWE["SW_DECOM"] = pd.to_datetime(
            SWE["SW_DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        current_ordered = SWE.sort_values(by=["SW_DECOM"], ascending=True)
        current_ordered = current_ordered.drop_duplicates(["CHILD"], keep="first")
        current_ordered["index"] = current_ordered.index

        no_prev_dec = prev_ordered[(prev_ordered["SW_DEC"].isna())]

        merged_df = no_prev_dec.merge(
            current_ordered, on="CHILD", how="left", suffixes=("_current", "_prev")
        )

        error_merged = merged_df[
            merged_df["SW_REASON_current"] != merged_df["SW_REASON_prev"]
        ]

        error_rows = SWE[SWE.index.isin(error_merged["index"])].index

        return {"SWEpisodes": error_rows.tolist()}


def test_validate():
    import pandas as pd

    fake_SWE_last = pd.DataFrame(
        [
            {
                "CHILD": "child1",
                "SW_DECOM": "01/01/2000",
                "SW_DEC": pd.NA,
                "SW_REASON": "OTHERS",
            },
            {
                "CHILD": "child1",
                "SW_DECOM": "02/01/2000",
                "SW_DEC": "01/01/2001",
                "SW_REASON": "OTHERS",
            },
            {
                "CHILD": "child2",
                "SW_DECOM": "01/01/2001",
                "SW_DEC": pd.NA,
                "SW_REASON": "OTHERS",
            },
            {
                "CHILD": "child3",
                "SW_DECOM": "01/01/2001",
                "SW_DEC": pd.NA,
                "SW_REASON": "OTHERS",
            },
            {
                "CHILD": "child4",
                "SW_DECOM": "01/01/2001",
                "SW_DEC": "02/01/2001",
                "SW_REASON": "OTHERS",
            },
        ]
    )

    fake_SWE = pd.DataFrame(
        [
            {
                "CHILD": "child1",
                "SW_DECOM": pd.NA,
                "SW_DEC": "01/01/2001",
                "SW_REASON": "OTHERS",
            },  # 0 pass, most recent prev has dec
            {
                "CHILD": "child2",
                "SW_DECOM": "01/01/2001",
                "SW_DEC": pd.NA,
                "SW_REASON": "OTHERS",
            },  # 1, pass, matching reason
            {
                "CHILD": "child3",
                "SW_DECOM": "01/01/2001",
                "SW_DEC": pd.NA,
                "SW_REASON": "FCONTA",
            },  # 2, fail, reason different
            {
                "CHILD": "child3",
                "SW_DECOM": "02/01/2001",
                "SW_DEC": pd.NA,
                "SW_REASON": "FCONTA",
            },  # 3, pass, not first episode in year
            {
                "CHILD": "child4",
                "SW_DECOM": "03/01/2001",
                "SW_DEC": "04/01/2001",
                "SW_REASON": "FCONTA",
            },  # 4 pass different reason but has dec
        ]
    )

    fake_dfs = {"SWEpisodes": fake_SWE, "SWEpisodes_last": fake_SWE_last}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="120",
    message="The reason for the reversal of the decision that the child should be placed for adoption code is not valid.",
    affected_fields=["REASON_PLACED_CEASED"],
    tables=["PlacedAdoption"],
)
def validate(dfs):
    if "PlacedAdoption" not in dfs:
        return {}

    placed_adoptions = dfs["PlacedAdoption"]
    code_list = ["RD1", "RD2", "RD3", "RD4"]

    mask = (
        placed_adoptions["REASON_PLACED_CEASED"].isin(code_list)
        | placed_adoptions["REASON_PLACED_CEASED"].isna()
    )

    validation_error_mask = ~mask
    validation_error_locations = placed_adoptions.index[validation_error_mask]

    return {"PlacedAdoption": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "REASON_PLACED_CEASED": ["rd1", "RD0", 1, "RD1", "", pd.NA],
        }
    )

    fake_dfs = {"PlacedAdoption": fake_data}

    result = validate(fake_dfs)

    assert result == {"PlacedAdoption": [0, 1, 2, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.rules.rule_utils import dec_after_decom


@rule_definition(
    code="376",
    message="Temporary placements coded as being due to holiday of usual foster carer(s) cannot exceed three weeks.",
    affected_fields=["DECOM", "PLACE"],
    tables=["Episodes", "Header"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Header" not in dfs:
        return {}
    else:
        return dec_after_decom(dfs, p_code="T3", y_gap=21)


def test_validate():
    fake_epi = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "444", "555", "666", "777"],
            "DECOM": [
                "01/01/2020",
                "01/02/2020",
                "02/03/2020",
                "15/03/2010",
                "15/03/2004",
                "14/03/2014",
                pd.NA,
            ],
            "DEC": [
                "22/01/2020",
                "23/02/2020",
                "24/03/2020",
                "06/04/2010",
                "05/04/2004",
                "04/04/2014",
                pd.NA,
            ],
            "PLACE": ["T3", "T3", "T3", "T3", "T3", "P2", "T3"],
        }
    )
    fake_hea = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "444", "555", "666"],
            "DOB": [
                "01/01/1998",
                "01/02/2000",
                "02/03/2015",
                "15/01/2010",
                "15/03/2000",
                "15/03/2000",
            ],
        }
    )

    fake_hea["DOB"] = pd.to_datetime(
        fake_hea["DOB"], format="%d/%m/%Y", errors="coerce"
    )
    fake_epi["DECOM"] = pd.to_datetime(
        fake_epi["DECOM"], format="%d/%m/%Y", errors="coerce"
    )
    fake_epi["DEC"] = pd.to_datetime(
        fake_epi["DEC"], format="%d/%m/%Y", errors="coerce"
    )

    fake_dfs = {"Episodes": fake_epi, "Header": fake_hea}

    assert validate(fake_dfs) == {"Episodes": [1, 2, 3]}



import pandas as pd
import numpy as np

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW15STG2",
    message="Reason for social worker change is missing.",
    affected_fields=["SW_REASON"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]
        collection_start = dfs["metadata"]["collection_start"]
        collection_start = pd.to_datetime(collection_start, format="%d/%m/%Y")

        df["SW_DECOM"] = pd.to_datetime(
            df["SW_DECOM"], format="%d/%m/%Y", errors="coerce"
        )

        reason_null = df[df["SW_REASON"].isna()]

        error_rows = reason_null[
            ~(reason_null["SW_DECOM"] < (collection_start - np.timedelta64(1, "Y")))
        ].index

        return {"SWEpisodes": error_rows.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "child1",
                "SW_DECOM": "02/01/1999",
                "SW_REASON": pd.NA,
            },  # 0 fail no reason and less than a year old
            {
                "CHILD": "child2",
                "SW_DECOM": "02/01/1998",
                "SW_REASON": pd.NA,
            },  # 1 pass, over a year old
            {
                "CHILD": "child3",
                "SW_DECOM": "02/01/1999",
                "SW_REASON": "OTHERS",
            },  # pass, less than a year old but reason given
        ]
    )

    fake_meta = {"collection_start": "01/01/2000"}

    fake_dfs = {"SWEpisodes": fake_data, "metadata": fake_meta}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [0]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="586",
    message="Dates of missing periods are before child’s date of birth.",
    affected_fields=["MIS_START"],
    tables=["Missing"],
)
def validate(dfs):
    if "Missing" not in dfs:
        return {}
    else:
        df = dfs["Missing"]
        df["DOB"] = pd.to_datetime(df["DOB"], format="%d/%m/%Y", errors="coerce")
        df["MIS_START"] = pd.to_datetime(
            df["MIS_START"], format="%d/%m/%Y", errors="coerce"
        )

        error_mask = df["MIS_START"].notna() & (df["MIS_START"] <= df["DOB"])
        return {"Missing": df.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"DOB": "01/02/2020", "MIS_START": pd.NA},  # 0
            {"DOB": "02/02/2020", "MIS_START": "07/02/2020"},  # 1
            {"DOB": "04/02/2020", "MIS_START": "03/02/2020"},  # 2  Fails
            {"DOB": "06/02/2020", "MIS_START": pd.NA},  # 3
            {"DOB": "07/02/2020", "MIS_START": "01/02/2020"},  # 4  Fails
            {"DOB": "08/02/2020", "MIS_START": "13/02/2020"},  # 5
        ]
    )

    fake_dfs = {"Missing": fake_data}

    assert validate(fake_dfs) == {"Missing": [2, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="432",
    message="The child ceased to be looked after at the end of the previous episode but the reason for "
    + "the new episode is not started to be looked after.",
    affected_fields=["RNE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]

        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")

        epi.sort_values(["CHILD", "DECOM"], inplace=True)
        epi.reset_index(inplace=True)
        epi.reset_index(inplace=True)

        rec_vals1 = ["E2", "E3", "E4A", "E4B", "E5", "E6", "E7", "E8", "E9", "E11"]
        rec_vals2 = [
            "E12",
            "E13",
            "E14",
            "E15",
            "E16",
            "E17",
            "E41",
            "E45",
            "E46",
            "E47",
            "E48",
        ]
        rec_vals = rec_vals1 + rec_vals2

        epi["LAG"] = epi["level_0"].shift(-1)

        err_co = epi.merge(
            epi,
            how="inner",
            left_on="level_0",
            right_on="LAG",
            suffixes=["", "_PRE"],
        ).query("CHILD == CHILD_PRE & REC_PRE.isin(@rec_vals) & RNE != 'S'")

        err_list = err_co["index"].unique().tolist()
        err_list.sort()

        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"CHILD": "111", "RNE": "P", "DECOM": "01/06/2020", "REC": "E2"},  # 0
            {"CHILD": "111", "RNE": "P", "DECOM": "05/06/2020", "REC": "E3"},  # 1
            {"CHILD": "111", "RNE": "T", "DECOM": "06/06/2020", "REC": "E3"},  # 2
            {"CHILD": "111", "RNE": "S", "DECOM": "08/06/2020", "REC": pd.NA},  # 3
            {"CHILD": "222", "RNE": "S", "DECOM": "05/06/2020", "REC": "E3"},  # 4
            {"CHILD": "333", "RNE": "B", "DECOM": "06/06/2020", "REC": "E3"},  # 5
            {"CHILD": "333", "RNE": "U", "DECOM": "10/06/2020", "REC": pd.NA},  # 6
            {"CHILD": "444", "RNE": "S", "DECOM": "07/06/2020", "REC": "E3"},  # 7
            {"CHILD": "444", "RNE": "S", "DECOM": "08/06/2020", "REC": "E3"},  # 8
            {"CHILD": "444", "RNE": "T", "DECOM": "09/06/2020", "REC": "E3"},  # 9
            {"CHILD": "444", "RNE": "S", "DECOM": "15/06/2020", "REC": "E3"},  # 10
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1, 2, 6, 9]}
